# import os
import sys
from typing import Generator

from sqlalchemy.dialects.postgresql import JSONB as POSTGRES_JSONB
from sqlalchemy.types import JSON

from mavedb.db.session import SessionLocal


def get_db() -> Generator:
    db = SessionLocal()
    db.current_user_id = None
    try:
        yield db
    finally:
        db.close()


# if 'PYTEST_RUN_CONFIG' in os.environ:
if "pytest" in sys.modules:
    JSONB = JSON
else:
    JSONB = POSTGRES_JSONB



from sqlalchemy import Column, Integer, String

from mavedb.db.base import Base


class AuthNonce(Base):
    __tablename__ = "social_auth_nonce"

    id = Column(Integer, primary_key=True, index=True)
    server_url = Column(String(255), nullable=False)
    timestamp = Column(Integer, nullable=False)
    salt = Column(String(65), nullable=False)



from typing import Optional

import numpy as np
import pandas as pd
from pandas.testing import assert_index_equal
from sqlalchemy import func, or_
from sqlalchemy.orm import Session

from mavedb.lib.array_comparison import assert_array_equal
from mavedb.lib.exceptions import ValidationError
from mavedb.lib.mave.constants import (
    HGVS_NT_COLUMN,
    HGVS_PRO_COLUMN,
    HGVS_SPLICE_COLUMN,
    VARIANT_COUNT_DATA,
    VARIANT_SCORE_DATA,
)
from mavedb.lib.mave.utils import is_csv_null
from mavedb.models.experiment import Experiment
from mavedb.models.keyword import Keyword
from mavedb.models.reference_genome import ReferenceGenome
from mavedb.models.reference_map import ReferenceMap
from mavedb.models.scoreset import Scoreset
from mavedb.models.target_gene import TargetGene
from mavedb.models.user import User
from mavedb.view_models.search import ScoresetsSearch

VariantData = dict[str, Optional[dict[str, dict]]]


def search_scoresets(db: Session, owner: Optional[User], search: ScoresetsSearch) -> list[Scoreset]:
    scoresets_query = db.query(Scoreset)  # \
    # .filter(Scoreset.private.is_(False))

    if owner is not None:
        scoresets_query = scoresets_query.filter(Scoreset.created_by_id == owner.id)

    if search.published is not None:
        if search.published:
            scoresets_query = scoresets_query.filter(Scoreset.published_date is not None)
        else:
            scoresets_query = scoresets_query.filter(Scoreset.published_date is None)

    if search.text:
        lower_search_text = search.text.lower()
        scoresets_query = scoresets_query.filter(
            or_(
                Scoreset.urn.contains(lower_search_text),
                Scoreset.title.contains(lower_search_text),
                Scoreset.short_description.contains(lower_search_text),
                Scoreset.abstract_text.contains(lower_search_text),
                Scoreset.target_gene.has(func.lower(TargetGene.name).contains(lower_search_text)),
                Scoreset.target_gene.has(func.lower(TargetGene.category).contains(lower_search_text)),
                Scoreset.keyword_objs.any(func.lower(Keyword.text).contains(lower_search_text))
                # TODO Add: ORGANISM_NAME UNIPROT, ENSEMBL, REFSEQ, LICENSE, plus TAX_ID if numeric
            )
        )

    if search.targets:
        scoresets_query = scoresets_query.filter(Scoreset.target_gene.has(TargetGene.name.in_(search.targets)))

    if search.target_organism_names:
        scoresets_query = scoresets_query.filter(
            Scoreset.target_gene.has(
                TargetGene.reference_maps.any(
                    ReferenceMap.genome.has(ReferenceGenome.organism_name.in_(search.target_organism_names))
                )
            )
        )

    if search.target_types:
        scoresets_query = scoresets_query.filter(Scoreset.target_gene.has(TargetGene.category.in_(search.target_types)))

    scoresets: list[Scoreset] = (
        scoresets_query.join(Scoreset.experiment).join(Scoreset.target_gene).order_by(Experiment.title).all()
    )
    if not scoresets:
        scoresets = []
    return scoresets  # filter_visible_scoresets(scoresets)


def filter_visible_scoresets(items: list[Scoreset]):
    # TODO Take the user into account.
    return filter(lambda item: not item.private, items or [])


def validate_datasets_define_same_variants(scores, counts):
    """
    Checks if two `pd.DataFrame` objects parsed from uploaded files
    define the same variants.

    Parameters
    ----------
    scores : `pd.DataFrame`
        Scores dataframe parsed from an uploaded scores file.
    counts : `pd.DataFrame`
        Scores dataframe parsed from an uploaded counts file.
    """
    # TODO First, confirm that the two dataframes have the same HGVS columns.
    print(scores.columns)
    print(counts.columns)
    try:
        if HGVS_NT_COLUMN in scores:
            assert_array_equal(
                scores[HGVS_NT_COLUMN].sort_values().values,
                counts[HGVS_NT_COLUMN].sort_values().values,
            )
        if HGVS_SPLICE_COLUMN in scores:
            assert_array_equal(
                scores[HGVS_SPLICE_COLUMN].sort_values().values,
                counts[HGVS_SPLICE_COLUMN].sort_values().values,
            )
        if HGVS_PRO_COLUMN in scores:
            assert_array_equal(
                scores[HGVS_PRO_COLUMN].sort_values().values,
                counts[HGVS_PRO_COLUMN].sort_values().values,
            )
    except AssertionError:
        raise ValidationError(
            "Your score and counts files do not define the same variants. "
            "Check that the hgvs columns in both files match."
        )


def create_variants_data(scores, counts=None, index_col=None) -> list[VariantData]:
    """
    Given two `defaultdict`s `score_map` and `count_map`, create an
    `OrderedDict` indexed by `hgvs_nt` where the keys are the attribute
    fields required to instantiate a `variant.models.Variant` instance.

    NOTE: Assumes that the dataframes are indexed by their primary columns,
    and that they define the same variants in both hgvs columns.

    Parameters
    ----------
    scores : Union[`pd.DataFrame`, str]
        Map indexed by the primary hgvs column inferred during validation.
        Map values are `dict` records where the key-pairs are column-value
        pairs inferred from the `scores` file uploaded during submission.
    counts : Union[`pd.DataFrame`, str] optional
        Map indexed by the primary hgvs column inferred during validation.
        Map values are `dict` records where the key-pairs are column-value
        pairs inferred from the `counts` file uploaded during submission.
    index_col : str
        Column to use as index, which is used when grouping rows between
        dataframes.

    Returns
    -------
    `list`
        Formatted records that can be used to create `variant.models.Variant`
        instances.
    """

    if isinstance(scores, str):
        scores = pd.read_json(scores, orient="records")
    if isinstance(counts, str):
        counts = pd.read_json(counts, orient="records")

    has_count_data = counts is not None and len(counts) > 0
    has_score_data = scores is not None and len(scores) > 0

    if index_col:
        scores.index = pd.Index(scores[index_col])
        if has_count_data:
            counts.index = pd.Index(counts[index_col])

    if not has_score_data:
        return []

    if has_count_data:
        assert_index_equal(scores.index.sort_values(), counts.index.sort_values())
        validate_datasets_define_same_variants(scores, counts)

    variants = []
    for (primary_hgvs, group) in scores.groupby(by=scores.index, sort=False):
        score_records = group.to_dict(orient="records")
        if has_count_data:
            count_records = counts[counts.index == primary_hgvs].to_dict(orient="records")
            assert len(score_records) == len(count_records)
        else:
            # Make duplicates to zip with self when no count data.
            count_records = [r.copy() for r in score_records]

        for (sr, cr) in zip(score_records, count_records):
            hgvs_nt = sr.pop(HGVS_NT_COLUMN)
            hgvs_splice = sr.pop(HGVS_SPLICE_COLUMN)
            hgvs_pro = sr.pop(HGVS_PRO_COLUMN)

            if is_csv_null(hgvs_nt) or hgvs_nt is np.NaN or hgvs_nt == "nan":
                hgvs_nt = None
            if is_csv_null(hgvs_splice) or hgvs_splice is np.NaN or hgvs_splice == "nan":
                hgvs_splice = None
            if is_csv_null(hgvs_pro) or hgvs_pro is np.NaN or hgvs_pro == "nan":
                hgvs_pro = None

            cr.pop(HGVS_NT_COLUMN)
            cr.pop(HGVS_SPLICE_COLUMN)
            cr.pop(HGVS_PRO_COLUMN)

            # JSON fields cannot store np.NaN values, so convert np.NaN to None.
            for key, value in sr.items():
                if is_csv_null(value) or value is np.NaN:
                    sr[key] = None
            if cr:
                for key, value in cr.items():
                    if is_csv_null(value) or value is np.NaN:
                        cr[key] = None

            data = {VARIANT_SCORE_DATA: sr, VARIANT_COUNT_DATA: {} if cr == sr else cr}
            variant = {
                HGVS_NT_COLUMN: hgvs_nt,
                HGVS_SPLICE_COLUMN: hgvs_splice,
                HGVS_PRO_COLUMN: hgvs_pro,
                "data": data,
            }
            variants.append(variant)

    return variants



from datetime import date
from sqlalchemy import Column, Date, ForeignKey, Integer, String
from sqlalchemy.orm import backref, relationship

from mavedb.db.base import Base
from .ensembl_identifier import EnsemblIdentifier
from .reference_map import ReferenceMap
from .refseq_identifier import RefseqIdentifier
from .uniprot_identifier import UniprotIdentifier
from .wild_type_sequence import WildTypeSequence

# TODO Reformat code without removing dependencies whose use is not detected.


class TargetGene(Base):
    __tablename__ = "target_genes"

    id = Column(Integer, primary_key=True, index=True)

    name = Column(String, nullable=False)
    category = Column(String, nullable=False)

    # ensembl_id_id = Column(Integer, ForeignKey('metadata_ensemblidentifier.id'), nullable=True)
    # ensembl_id_id = Column(Integer, ForeignKey('ensembl_identifiers.id'), nullable=True)
    # ensembl_id = relationship('EnsemblIdentifier', backref='target_genes')
    # refseq_id_id = Column(Integer, nullable=True)  # , ForeignKey('dataset_scoreset.id'), nullable=False)
    scoreset_id = Column(Integer, ForeignKey("scoresets.id"), nullable=False)
    scoreset = relationship(
        "Scoreset",
        backref=backref("target_gene", cascade="all,delete-orphan", single_parent=True, uselist=False),
        single_parent=True,
    )
    # uniprot_id_id = Column(Integer, nullable=True)  # , ForeignKey('dataset_scoreset.id'), nullable=False)
    wt_sequence_id = Column(Integer, ForeignKey("wild_type_sequences.id"), nullable=False)
    wt_sequence = relationship(
        "WildTypeSequence",
        backref=backref("target_gene", single_parent=True, uselist=False),
        cascade="all,delete-orphan",
        single_parent=True,
    )

    creation_date = Column(Date, nullable=False, default=date.today)
    modification_date = Column(Date, nullable=False, default=date.today, onupdate=date.today)



from typing import Any, Optional

from fastapi import APIRouter

from mavedb import __project__, __version__
from mavedb.view_models import api_version

router = APIRouter(prefix="/api/v1/api", tags=["api information"], responses={404: {"description": "Not found"}})


class TestException(Exception):
    def __init__(self, message: Optional[str] = None):
        self.message = message


@router.get("/version", status_code=200, response_model=api_version.ApiVersion, responses={404: {}})
def show_version() -> Any:
    """
    Describe the API version.
    """

    return api_version.ApiVersion(
        name=__project__,
        version=__version__
    )


@router.get("/generateError", status_code=200, responses={404: {}})
def generate_error() -> Any:
    """
    Raise an exception that will not be caught except by FastAPI.
    """

    raise TestException("This error was generated for testing purposes.")



# valid data usage policies

import re

"""
Null Constant definitions
"""
NA_STRING = "NA"
null_values_list = (
    "nan",
    "na",
    "none",
    "",
    "undefined",
    "n/a",
    "null",
    "nil",
    "-",
    None,
)
# enforce the assumption that these are all lowercase values
null_values_list = [s.lower() for s in null_values_list if s is not None]
# add the NA_STRING only if it's not already in the list
if NA_STRING.lower() not in null_values_list:
    null_values_list.append(NA_STRING.lower())
null_values_list.sort()

null_values_re = re.compile(
    r"^\s+$|" + "|".join(f"^{s}$" for s in null_values_list if len(s)),
    flags=re.IGNORECASE,
)

readable_null_values_list = [f"{s}" for s in null_values_list] + ["whitespace"]

hgvs_nt_column = "hgvs_nt"
hgvs_splice_column = "hgvs_splice"
hgvs_pro_column = "hgvs_pro"
hgvs_columns = sorted([hgvs_nt_column, hgvs_pro_column, hgvs_splice_column])
meta_data = "meta_data"
score_columns = "score_columns"
count_columns = "count_columns"
variant_score_data = "score_data"
variant_count_data = "count_data"
required_score_column = "score"

valid_dataset_columns = [score_columns, count_columns]
valid_variant_columns = [variant_score_data, variant_count_data]

variant_to_scoreset_column = {
    variant_score_data: score_columns,
    variant_count_data: count_columns,
}
scoreset_to_variant_column = {v: k for k, v in variant_to_scoreset_column.items()}



from mavedb.lib.validation.exceptions import ValidationError
from mavedb.lib.validation.constants.target import valid_categories, valid_sequence_types
from fqfa import infer_sequence_type
from fqfa.validator import amino_acids_validator, dna_bases_validator


def validate_target_category(category: str):
    """
    If the target category provided does not fall within a pre-defined list of valid categories.

    Parameters
    __________
    category: str
        The target category to be validated.

    Raises
    ______
    ValidationError
        If the target category provided is not valid.
    """
    if category not in valid_categories:
        raise ValidationError(
            "{} is not a valid target category. Valid categories are "
            "Protein coding, Regulatory, and Other noncoding".format(category)
        )


def validate_sequence_category(sequence_type: str):
    """
    If the sequence type provided does not fall within a pre-defined list of valid sequence types.

    Parameters
    __________
    sequence_type: str
        The sequence type to be validated.

    Raises
    ______
    ValidationError
        If the sequence type provided is not valid.
    """
    if sequence_type not in valid_sequence_types:
        raise ValidationError(f"'{sequence_type}' is not a valid sequence type")


def validate_target_sequence(target_seq: str, target_seq_type: str):
    """
    Validates a target sequence whether match sequence_type.

    Parameters
    __________
    sequence : str
        The target sequence that will be validated.

    Raises
    ______
    ValidationError
        If the target sequence does not consist of ACTG or Amino acid.
    """
    if target_seq_type == "infer":
        if infer_sequence_type(target_seq) not in valid_sequence_types:
            raise ValidationError(f"invalid inferred sequence type '{infer_sequence_type(target_seq)}'")
    elif target_seq_type == "dna":
        if dna_bases_validator(target_seq) is None:
            raise ValidationError("invalid dna sequence provided")
    elif target_seq_type == "protein":
        if amino_acids_validator(target_seq) is None:
            raise ValidationError("invalid protein sequence provided")
    else:
        raise ValueError(f"unexpected sequence type '{target_seq_type}'")



from datetime import date
from typing import Optional

from mavedb.view_models.base.base import BaseModel
from mavedb.view_models.reference_genome import ReferenceGenome, ReferenceGenomeInDbBase


class ReferenceMapBase(BaseModel):
    # organism_name_id: int
    pass


class ReferenceMapCreate(ReferenceMapBase):
    genome_id: Optional[int]
    target_id: Optional[int]


# Properties shared by models stored in DB
class SavedReferenceMap(ReferenceMapBase):
    id: int
    genome_id: int
    target_id: int
    is_primary: bool
    genome: ReferenceGenomeInDbBase
    creation_date: date
    modification_date: date

    class Config:
        orm_mode = True


# Properties to return to client
class ReferenceMap(SavedReferenceMap):
    genome: ReferenceGenome



import logging
from operator import or_
from typing import Optional

from sqlalchemy.orm import Session

from mavedb.models.experiment import Experiment
from mavedb.models.scoreset import Scoreset
from mavedb.models.user import User
from mavedb.view_models.search import ExperimentsSearch

logger = logging.getLogger(__name__)


def search_experiments(db: Session, owner: Optional[User], search: ExperimentsSearch) -> list[Scoreset]:
    query = db.query(Experiment)
    # .filter(Scoreset.private.is_(False))

    if owner is not None:
        query = query.filter(Experiment.created_by_id == owner.id)

    if search.published is not None:
        if search.published:
            query = query.filter(Experiment.published_date is not None)
        else:
            query = query.filter(Experiment.published_date is None)

    if search.text:
        lower_search_text = search.text.lower()
        query = query.filter(
            or_(
                Experiment.urn.contains(lower_search_text),
                Experiment.title.contains(lower_search_text),
                Experiment.short_description.contains(lower_search_text),
                Experiment.abstract_text.contains(lower_search_text),
            )
        )

    items: list[Experiment] = query.order_by(Experiment.title).all()
    if not items:
        items = []
    return items



from unittest import TestCase

from mavedb.lib.validation.identifier import (
    validate_sra_identifier,
    validate_sra_list,
    validate_uniprot_identifier,
    validate_refseq_identifier,
    validate_refseq_list,
    validate_genome_identifier,
    validate_ensembl_identifier,
    validate_ensembl_list,
)
from mavedb.lib.validation.exceptions import ValidationError


class TestGenomeValidators(TestCase):
    """
    Tests that each validator throws the appropriate :class:`ValidationError`
    when passed invalid Genome assemblies as well as detecting correct ones.
    """

    def test_ve_invalid(self):
        with self.assertRaises(ValidationError):
            validate_genome_identifier("GCF_000146045")

    def test_passes_valid(self):
        validate_genome_identifier("GCF_000001405.11")
        validate_genome_identifier("GCA_000001405.11")


class TestSRAValidators(TestCase):
    """
    Tests that each validator throws the appropriate :class:`ValidationError`
    when passed invalid SRAs as well as detecting correct SRAs
    """

    def test_ve_invalid_sra(self):
        with self.assertRaises(ValidationError):
            validate_sra_identifier("SRXP0001")

    def test_ve_invalid_sra_in_list(self):
        with self.assertRaises(ValidationError):
            validate_sra_list(["SRX3407686", "SRXPPPPP0001"])

    def test_passes_valid_id(self):
        validate_sra_identifier("SRX3407687")


class TestEnsemblValidators(TestCase):
    """
    Tests that each validator throws the appropriate :class:`ValidationError`
    when passed invalid Ensembl as well as detecting correct Ensembl ids.
    """

    def test_ve_invalid_id(self):
        with self.assertRaises(ValidationError):
            validate_ensembl_identifier("ENVS00000010404")

    def test_ve_invalid_list(self):
        with self.assertRaises(ValidationError):
            validate_ensembl_list(["ENSG00000143384", "NO_155436.1"])

    def test_passes_valid_id(self):
        validate_ensembl_identifier("ENSG00000143384")


class TestRefSeqValidators(TestCase):
    """
    Tests that each validator throws the appropriate :class:`ValidationError`
    when passed invalid RefSeq as well as detecting correct RefSeq ids.
    """

    def test_ve_invalid_id(self):
        with self.assertRaises(ValidationError):
            validate_refseq_identifier("NX_155436.1")

    def test_ve_invalid_list(self):
        with self.assertRaises(ValidationError):
            validate_refseq_list(["NR_155436.1", "NO_155436.1"])

    def test_passes_valid_id(self):
        validate_refseq_identifier("NR_155436.1")


class TestUniprotValidators(TestCase):
    """
    Tests that each validator throws the appropriate :class:`ValidationError`
    when passed invalid Uniprot as well as detecting correct Uniprot ids.
    """

    def test_ve_invalid_uniprot_id(self):
        with self.assertRaises(ValidationError):
            validate_uniprot_identifier("P123")

    def test_passes_valid_uniprot_id(self):
        validate_uniprot_identifier("P01133")



from fastapi.testclient import TestClient
import os
import pytest
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from mavedb.server_main import app
from mavedb.db.base import Base
from mavedb.deps import get_db
from mavedb.lib.authorization import require_current_user
from mavedb.models.user import User


# To store the test database temporarily on disk:
SQLITE_DB_FILE = "./test.db"

# To store the test database in memory:
# SQLITE_DB_FILE = None

SQLALCHEMY_DATABASE_URL = "sqlite://" if SQLITE_DB_FILE is None else f"sqlite:///{SQLITE_DB_FILE}"

engine = create_engine(SQLALCHEMY_DATABASE_URL, connect_args={"check_same_thread": False})
TestingSessionLocal = sessionmaker(autocommit=False, autoflush=False, bind=engine)


def override_get_db():
    try:
        db = TestingSessionLocal()
        yield db
    finally:
        db.close()


def override_current_user():
    yield User(
        # id=1,
        username="someuser",
        first_name="First",
        last_name="Last",
        is_active=True,
        is_staff=False,
        is_superuser=False,
    )


app.dependency_overrides[get_db] = override_get_db
app.dependency_overrides[require_current_user] = override_current_user

client = TestClient(app)


@pytest.fixture()
def test_empty_db():
    Base.metadata.create_all(bind=engine)
    yield
    Base.metadata.drop_all(bind=engine)
    if SQLITE_DB_FILE is not None:
        os.remove(SQLITE_DB_FILE)


@pytest.fixture()
def test_db_with_dataset():
    Base.metadata.create_all(bind=engine)
    yield
    Base.metadata.drop_all(bind=engine)



from typing import Optional

from fastapi import Depends, HTTPException
from starlette import status

from mavedb.lib.authentication import get_current_user
from mavedb.models.user import User

ROLES = {"admin": "admin"}


####################################################################################################
# Main authorization methods
####################################################################################################


async def require_current_user(user: Optional[User] = Depends(get_current_user)) -> User:
    if user is None:
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED, detail="Could not validate credentials")
    user_value: User = user
    return user_value


class RoleRequirer:
    def __init__(self, roles: list[str]):
        self.roles = roles

    async def __call__(self, user: User = Depends(require_current_user)):
        if not any(x in user.roles for x in self.roles):
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED, detail="You are not authorized to use this feature"
            )
        return user


async def require_role(roles: list[str], user: User = Depends(require_current_user)) -> User:
    if not any(x in user.roles for x in roles):
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED, detail="You are not authorized to use this feature"
        )
    return user



from datetime import date
from typing import Optional

from mavedb.view_models.base.base import BaseModel


class LicenseBase(BaseModel):
    """Base class for license view models."""
    long_name: str
    short_name: str
    link: Optional[str]
    version: Optional[str]


# Properties shared by models stored in DB
class SavedLicense(LicenseBase):
    """Base class for license view models representing saved records."""

    id: int

    class Config:
        orm_mode = True


class ShortLicense(SavedLicense):
    """License view model containing a smaller set of properties to return in list contexts."""

    pass


# Properties to return to non-admin clients
class License(SavedLicense):
    """License view model containing properties visible to all users."""

    text: str
    creation_date: date
    modification_date: date



from typing import Any

from fastapi import APIRouter, Depends, HTTPException
from sqlalchemy.exc import MultipleResultsFound
from sqlalchemy.orm import Session

from mavedb import deps
from mavedb.lib.authorization import require_current_user, RoleRequirer
from mavedb.models.user import User
from mavedb.view_models import user

router = APIRouter(prefix="/api/v1", tags=["access keys"], responses={404: {"description": "Not found"}})


@router.get("/users/", status_code=200, response_model=list[user.AdminUser], responses={404: {}})
async def list_users(*, db: Session = Depends(deps.get_db), user: User = Depends(RoleRequirer(["admin"]))) -> Any:
    """
    List users.
    """
    items = db.query(User).order_by(User.username).all()
    return items


@router.get("/users/me", status_code=200, response_model=user.CurrentUser, responses={404: {}, 500: {}})
async def show_me(*, user: User = Depends(require_current_user)) -> Any:
    """
    Return the current user.
    """
    return user


@router.get("/users/{id}", status_code=200, response_model=user.AdminUser, responses={404: {}, 500: {}})
async def show_user(
    *, id: int, user: User = Depends(RoleRequirer(["admin"])), db: Session = Depends(deps.get_db)
) -> Any:
    """
    Fetch a single user by ID.
    """
    try:
        item = db.query(User).filter(User.id == id).one_or_none()
    except MultipleResultsFound:
        raise HTTPException(status_code=500, detail=f"Multiple users with ID {id} were found.")
    if not item:
        raise HTTPException(status_code=404, detail=f"User with ID {id} not found")
    return item


@router.put("/users/me", status_code=200, response_model=user.CurrentUser, responses={404: {}, 500: {}})
async def update_me(
    *, user_update: user.CurrentUserUpdate, db: Session = Depends(deps.get_db), user: User = Depends(require_current_user)
) -> Any:
    """
    Update the current user.
    """
    user.email = user_update.email
    db.add(user)
    db.commit()
    db.refresh(user)
    return user


@router.put("/users//{id}", status_code=200, response_model=user.AdminUser, responses={404: {}, 500: {}})
async def update_user(
    *,
    id: int,
    item_update: user.AdminUserUpdate,
    db: Session = Depends(deps.get_db),
    user: User = Depends(require_current_user),
) -> Any:
    """
    Update a user.
    """
    if not item_update:
        raise HTTPException(status_code=400, detail="The request contained no updated item.")

    try:
        item = db.query(User).filter(User.id == id).one_or_none()
    except MultipleResultsFound:
        raise HTTPException(status_code=500, detail=f"Multiple users with ID {id} were found.")
    if not item:
        raise HTTPException(status_code=404, detail=f"User with id {id} not found.")
    # TODO Ensure that the current user has edit rights for this scoreset.

    item.first_name = item_update.first_name
    item.last_name = item_update.last_name
    item.email = item_update.email
    await item.set_roles(db, item_update.roles)
    db.add(item)
    db.commit()
    db.refresh(item)
    return item



from typing import Any, List

from fastapi import APIRouter, Depends, HTTPException
from sqlalchemy.orm import Session

from mavedb import deps
from mavedb.models.license import License
from mavedb.view_models import license

router = APIRouter(prefix="/api/v1/licenses", tags=["licenses"], responses={404: {"description": "Not found"}})


@router.get("/", status_code=200, response_model=List[license.ShortLicense], responses={404: {}})
def list_licenses(
    *,
    db: Session = Depends(deps.get_db),
) -> Any:
    """
    List licenses.
    """

    items = db.query(License).order_by(License.short_name).all()
    return items


@router.get("/{item_id}", status_code=200, response_model=license.License, responses={404: {}})
def fetch_license(
    *,
    item_id: int,
    db: Session = Depends(deps.get_db),
) -> Any:
    """
    Fetch a single license by ID.
    """

    item = db.query(License).filter(License.id == item_id).first()
    if not item:
        raise HTTPException(status_code=404, detail=f"License with ID {item_id} not found")
    return item



from typing import Any, List

from fastapi import APIRouter, Depends, HTTPException
from sqlalchemy import func
from sqlalchemy.orm import Session

from mavedb import deps
from mavedb.models.target_gene import TargetGene
from mavedb.view_models import target_gene
from mavedb.view_models.search import TextSearch

router = APIRouter(prefix="/api/v1/targetGenes", tags=["target-genes"], responses={404: {"description": "Not found"}})


@router.get("/", status_code=200, response_model=List[target_gene.TargetGene], responses={404: {}})
def list_target_genes(
    *,
    db: Session = Depends(deps.get_db),
) -> Any:
    """
    List target genes.
    """
    items = db.query(TargetGene).order_by(TargetGene.name).all()
    return items


@router.get("/names", status_code=200, response_model=List[str], responses={404: {}})
def list_target_gene_names(
    *,
    db: Session = Depends(deps.get_db),
) -> Any:
    """
    List distinct target gene names, in alphabetical order.
    """

    items = db.query(TargetGene).all()
    names = map(lambda item: item.name, items)
    return sorted(list(set(names)))


@router.get("/categories", status_code=200, response_model=List[str], responses={404: {}})
def list_target_gene_categories(
    *,
    db: Session = Depends(deps.get_db),
) -> Any:
    """
    List distinct target genes categories, in alphabetical order.
    """

    items = db.query(TargetGene).all()
    categories = map(lambda item: item.category, items)
    return sorted(list(set(categories)))


@router.get("/{item_id}", status_code=200, response_model=target_gene.TargetGene, responses={404: {}})
def fetch_target_gene(
    *,
    item_id: int,
    db: Session = Depends(deps.get_db),
) -> Any:
    """
    Fetch a single target gene by ID.
    """
    item = db.query(TargetGene).filter(TargetGene.id == item_id).first()
    if not item:
        raise HTTPException(status_code=404, detail=f"TargetGene with ID {item_id} not found")
    return item


@router.post("/search", status_code=200, response_model=List[target_gene.TargetGene])
def search_target_genes(search: TextSearch, db: Session = Depends(deps.get_db)) -> Any:
    """
    Search target genes.
    """

    query = db.query(TargetGene)

    if search.text and len(search.text.strip()) > 0:
        lower_search_text = search.text.strip().lower()
        query = query.filter(func.lower(TargetGene.name).contains(lower_search_text))
    else:
        raise HTTPException(status_code=500, detail="Search text is required")

    items = query.order_by(TargetGene.name).limit(50).all()
    if not items:
        items = []
    return items



from sqlalchemy import Column, Integer, ForeignKey
from sqlalchemy.orm import relationship, backref

from mavedb.db.base import Base


class RefseqOffset(Base):
    __tablename__ = "refseq_offsets"

    identifier_id = Column(Integer, ForeignKey("refseq_identifiers.id"), nullable=False, primary_key=True)
    identifier = relationship("RefseqIdentifier", backref=backref("target_gene_offsets", uselist=True))
    target_gene_id = Column(Integer, ForeignKey("target_genes.id"), nullable=False, primary_key=True)
    target_gene = relationship(
        "TargetGene",
        backref=backref("refseq_offset", cascade="all,delete-orphan", single_parent=True, uselist=False),
        single_parent=True,
    )
    offset = Column(Integer, nullable=False)



# Datasets
VARIANT_SCORE_DATA = "score_data"
VARIANT_COUNT_DATA = "count_data"

# Dataset columns
HGVS_NT_COLUMN = "hgvs_nt"
HGVS_SPLICE_COLUMN = "hgvs_splice"
HGVS_PRO_COLUMN = "hgvs_pro"
HGVS_COLUMNS = sorted([HGVS_NT_COLUMN, HGVS_PRO_COLUMN, HGVS_SPLICE_COLUMN])

# META_DATA = 'meta_data'

SCORE_COLUMNS = "score_columns"
COUNT_COLUMNS = "count_columns"

REQUIRED_SCORE_COLUMN = "score"



from typing import Union

import metapub
from eutils import EutilsNCBIError
from sqlalchemy.orm import Session

from mavedb.models.doi_identifier import DoiIdentifier
from mavedb.models.ensembl_identifier import EnsemblIdentifier
from mavedb.models.ensembl_offset import EnsemblOffset
from mavedb.models.pubmed_identifier import PubmedIdentifier
from mavedb.models.refseq_identifier import RefseqIdentifier
from mavedb.models.refseq_offset import RefseqOffset
from mavedb.models.target_gene import TargetGene
from mavedb.models.uniprot_identifier import UniprotIdentifier
from mavedb.models.uniprot_offset import UniprotOffset
from mavedb.models.raw_read_identifier import RawReadIdentifier

EXTERNAL_GENE_IDENTIFIER_CLASSES = {
    "Ensembl": EnsemblIdentifier,
    "RefSeq": RefseqIdentifier,
    "UniProt": UniprotIdentifier,
}

EXTERNAL_GENE_IDENTIFIER_OFFSET_CLASSES = {"Ensembl": EnsemblOffset, "RefSeq": RefseqOffset, "UniProt": UniprotOffset}

EXTERNAL_GENE_IDENTIFIER_OFFSET_ATTRIBUTES = {
    "Ensembl": "ensembl_offset",
    "RefSeq": "refseq_offset",
    "UniProt": "uniprot_offset",
}


async def find_or_create_doi_identifier(db: Session, identifier: str):
    """
    Find an existing DOI identifier record with the specified identifier string, or create a new one.

    :param db: An active database session
    :param identifier: A valid DOI identifier
    :return: An existing DoiIdentifier containing the specified identifier string, or a new, unsaved DoiIdentifier
    """
    doi_identifier = db.query(DoiIdentifier).filter(DoiIdentifier.identifier == identifier).one_or_none()
    if not doi_identifier:
        doi_identifier = DoiIdentifier(identifier=identifier, db_name="DOI", url=f"https://doi.org/{identifier}")
    return doi_identifier


def fetch_pubmed_citation_html(identifier: str):
    fetch = metapub.PubMedFetcher()
    try:
        article = fetch.article_by_pmid(identifier)
    except EutilsNCBIError:
        return f"Unable to retrieve PubMed ID {identifier}"
    else:
        return article.citation_html


async def find_or_create_pubmed_identifier(db: Session, identifier: str):
    """
    Find an existing PubMed identifier record with the specified identifier string, or create a new one.

    :param db: An active database session
    :param identifier: A valid PubMed identifier
    :return: An existing PubmedIdentifier containing the specified identifier string, or a new, unsaved PubmedIdentifier
    """
    pubmed_identifier = db.query(PubmedIdentifier).filter(PubmedIdentifier.identifier == identifier).one_or_none()
    if not pubmed_identifier:
        pubmed_identifier = PubmedIdentifier(
            identifier=identifier,
            db_name="PubMed",
            url=f"http://www.ncbi.nlm.nih.gov/pubmed/{identifier}",
            reference_html=fetch_pubmed_citation_html(identifier),
        )
    return pubmed_identifier


async def find_or_create_raw_read_identifier(db: Session, identifier: str):
    raw_read_identifier = db.query(RawReadIdentifier).filter(RawReadIdentifier.identifier == identifier).one_or_none()
    if not raw_read_identifier:
        raw_read_identifier = RawReadIdentifier(
            identifier=identifier, db_name="SRA", url=f"http://www.ebi.ac.uk/ena/data/view/{identifier}"
        )
    return raw_read_identifier


async def find_or_create_external_gene_identifier(db: Session, db_name: str, identifier: str):
    """
    Find an existing gene identifier record with the specified gene database name and identifier string, or create a new
    one.

    :param db: An active database session
    :param identifier: A valid identifier
    :return: An existing EnsemblIdentifier, RefseqIdentifier, or UniprotIdentifier containing the specified identifier
      string, or a new, unsaved instance of one of these classes
    """

    # TODO Handle key errors.
    identifier_class: Union[EnsemblIdentifier, RefseqIdentifier, UniprotIdentifier] = EXTERNAL_GENE_IDENTIFIER_CLASSES[
        db_name
    ]

    external_gene_identifier = (
        db.query(identifier_class).filter(identifier_class.identifier == identifier).one_or_none()
    )

    if not external_gene_identifier:
        external_gene_identifier = identifier_class(
            identifier=identifier,
            db_name=db_name
            # TODO Set URL from identifier
            # url=f'https://doi.org/{identifier}'
        )
    return external_gene_identifier


async def create_external_gene_identifier_offset(
    db: Session, target_gene: TargetGene, db_name: str, identifier: str, offset: int
):
    external_gene_identifier = await find_or_create_external_gene_identifier(db, db_name, identifier)

    # TODO Handle key errors.
    offset_class = EXTERNAL_GENE_IDENTIFIER_OFFSET_CLASSES[external_gene_identifier.db_name]
    external_gene_identifier_offset = offset_class(identifier=external_gene_identifier, offset=offset)

    identifier_offset_attribute = EXTERNAL_GENE_IDENTIFIER_OFFSET_ATTRIBUTES[external_gene_identifier.db_name]
    setattr(target_gene, identifier_offset_attribute, external_gene_identifier_offset)



import os
from datetime import date
import json
import re

from mavedb.models.reference_genome import ReferenceGenome
from tests.conftest import client, TestingSessionLocal

"""
#Test one. Can ignore it.
def test_get_scoreset(test_empty_db):
    response = client.get("/api/v1/scoresets/")
    assert response.status_code == 200
    response_data = response.json()
    scoreset_urn = response_data['urn']
    assert scoreset_urn is not None
    assert re.match(r'tmp:[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}', scoreset_urn)
"""


def create_an_experiment():
    experiment_to_create = {
        "title": "Test Experiment Title",
        "methodText": "Methods",
        "abstractText": "Abstract",
        "shortDescription": "Test experiment",
        "extraMetadata": {"key": "value"},
        "keywords": [],
    }
    response = client.post("/api/v1/experiments/", json=experiment_to_create)
    experiment = response.json()
    # experiment_urn = response_data['urn']
    # experiment_set_urn = response_data['experimentSetUrn']
    return experiment


def create_reference_genome():
    reference_genome = ReferenceGenome(id=1, short_name="Name", organism_name="Organism")
    db = TestingSessionLocal()
    db.add(reference_genome)
    db.commit()


def create_scoreset():
    experiment = create_an_experiment()
    scoreset_to_create = {
        "experimentUrn": experiment["urn"],
        "title": "Test Scoreset Title",
        "methodText": "Methods",
        "abstractText": "Abstract",
        "shortDescription": "Test scoreset",
        "extraMetadata": {"key": "value"},
        "keywords": [],
        "doiIdentifiers": [],
        "pubmedIdentifiers": [],
        "dataUsagePolicy": None,
        "datasetColumns": {},
        "supersededScoresetUrn": None,
        "metaAnalysisSourceScoresetUrns": [],
        "targetGene": {
            "name": "UBE2I",
            "category": "Protein coding",
            "externalIdentifiers": [
                {"identifier": {"dbName": "Ensembl", "identifier": "ENSG00000103275"}, "offset": 0},
                {"identifier": {"dbName": "RefSeq", "identifier": "NM_003345"}, "offset": 159},
                {"identifier": {"dbName": "UniProt", "identifier": "P63279"}, "offset": 0},
            ],
            "referenceMaps": [{"genomeId": 1}],
            "wtSequence": {
                "sequenceType": "dna",
                "sequence": "ATGAGTATTCAACATTTCCGTGTCGCCCTTATTCCCTTTTTTGCGGCATTTTGCCTTCCTGTTTTTGCTCACCCAGAAACGCTGGTGAAAGTAAAAGATGCTGAAGATCAGTTGGGTGCACGAGTGGGTTACATCGAACTGGATCTCAACAGCGGTAAGATCCTTGAGAGTTTTCGCCCCGAAGAACGTTTTCCAATGATGAGCACTTTTAAAGTTCTGCTATGTGGCGCGGTATTATCCCGTGTTGACGCCGGGCAAGAGCAACTCGGTCGCCGCATACACTATTCTCAGAATGACTTGGTTGAGTACTCACCAGTCACAGAAAAGCATCTTACGGATGGCATGACAGTAAGAGAATTATGCAGTGCTGCCATAACCATGAGTGATAACACTGCGGCCAACTTACTTCTGACAACGATCGGAGGACCGAAGGAGCTAACCGCTTTTTTGCACAACATGGGGGATCATGTAACTCGCCTTGATCGTTGGGAACCGGAGCTGAATGAAGCCATACCAAACGACGAGCGTGACACCACGATGCCTGCAGCAATGGCAACAACGTTGCGCAAACTATTAACTGGCGAACTACTTACTCTAGCTTCCCGGCAACAATTAATAGACTGGATGGAGGCGGATAAAGTTGCAGGACCACTTCTGCGCTCGGCCCTTCCGGCTGGCTGGTTTATTGCTGATAAATCTGGAGCCGGTGAGCGTGGGTCTCGCGGTATCATTGCAGCACTGGGGCCAGATGGTAAGCCCTCCCGTATCGTAGTTATCTACACGACGGGGAGTCAGGCAACTATGGATGAACGAAATAGACAGATCGCTGAGATAGGTGCCTCACTGATTAAGCATTGGTAA",
            },
        },
    }
    response = client.post("/api/v1/scoresets/", json=scoreset_to_create)
    scoreset = response.json()
    return scoreset


def test_create_scoreset_with_some_none_values(test_empty_db):
    experiment = create_an_experiment()
    create_reference_genome()
    scoreset_to_create = {
        "experimentUrn": experiment["urn"],
        "title": "Test Scoreset Title",
        "methodText": "Methods",
        "abstractText": "Abstract",
        "shortDescription": "Test scoreset",
        "extraMetadata": {"key": "value"},
        "keywords": [],
        "doiIdentifiers": [],
        "pubmedIdentifiers": [],
        "dataUsagePolicy": None,
        "datasetColumns": {},
        "supersededScoresetUrn": None,
        "metaAnalysisSourceScoresetUrns": [],
        "targetGene": {
            "name": "UBE2I",
            "category": "Protein coding",
            "externalIdentifiers": [
                {"identifier": {"dbName": "Ensembl", "identifier": "ENSG00000103275"}, "offset": 0},
                {"identifier": {"dbName": "RefSeq", "identifier": "NM_003345"}, "offset": 159},
                {"identifier": {"dbName": "UniProt", "identifier": "P63279"}, "offset": 0},
            ],
            "referenceMaps": [{"genomeId": 1}],
            "wtSequence": {
                "sequenceType": "dna",
                "sequence": "ATGAGTATTCAACATTTCCGTGTCGCCCTTATTCCCTTTTTTGCGGCATTTTGCCTTCCTGTTTTTGCTCACCCAGAAACGCTGGTGAAAGTAAAAGATGCTGAAGATCAGTTGGGTGCACGAGTGGGTTACATCGAACTGGATCTCAACAGCGGTAAGATCCTTGAGAGTTTTCGCCCCGAAGAACGTTTTCCAATGATGAGCACTTTTAAAGTTCTGCTATGTGGCGCGGTATTATCCCGTGTTGACGCCGGGCAAGAGCAACTCGGTCGCCGCATACACTATTCTCAGAATGACTTGGTTGAGTACTCACCAGTCACAGAAAAGCATCTTACGGATGGCATGACAGTAAGAGAATTATGCAGTGCTGCCATAACCATGAGTGATAACACTGCGGCCAACTTACTTCTGACAACGATCGGAGGACCGAAGGAGCTAACCGCTTTTTTGCACAACATGGGGGATCATGTAACTCGCCTTGATCGTTGGGAACCGGAGCTGAATGAAGCCATACCAAACGACGAGCGTGACACCACGATGCCTGCAGCAATGGCAACAACGTTGCGCAAACTATTAACTGGCGAACTACTTACTCTAGCTTCCCGGCAACAATTAATAGACTGGATGGAGGCGGATAAAGTTGCAGGACCACTTCTGCGCTCGGCCCTTCCGGCTGGCTGGTTTATTGCTGATAAATCTGGAGCCGGTGAGCGTGGGTCTCGCGGTATCATTGCAGCACTGGGGCCAGATGGTAAGCCCTCCCGTATCGTAGTTATCTACACGACGGGGAGTCAGGCAACTATGGATGAACGAAATAGACAGATCGCTGAGATAGGTGCCTCACTGATTAAGCATTGGTAA",
            },
        },
    }
    response = client.post("/api/v1/scoresets/", json=scoreset_to_create)
    assert response.status_code == 200
    response_data = response.json()
    scoreset_urn = response_data["urn"]
    experiment_urn = experiment["urn"]
    assert scoreset_urn is not None
    assert experiment_urn is not None
    assert re.match(r"tmp:[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}", scoreset_urn)
    assert re.match(r"tmp:[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}", experiment_urn)
    # Don't use same way as test_create_experiment because some keys in submitted scoreset are different from the
    # response_data. For example, in submitted scoreset experiment urn is called experimentUrn while in response_data
    # it's in experiment{} and called urn.
    # More data are added in different sub-dict such as externalIdentifiers, referenceMaps
    expected_response = {
        "urn": scoreset_urn,
        "title": "Test Scoreset Title",
        "methodText": "Methods",
        "abstractText": "Abstract",
        "shortDescription": "Test scoreset",
        "extraMetadata": {"key": "value"},
        "dataUsagePolicy": None,
        "licenceId": None,
        "keywords": [],
        "numVariants": 0,
        "experiment": {
            "title": "Test Experiment Title",
            "shortDescription": "Test experiment",
            "abstractText": "Abstract",
            "methodText": "Methods",
            "extraMetadata": {"key": "value"},
            "keywords": [],
            "urn": experiment_urn,
            "numScoresets": 0,
            "createdBy": {"orcidId": "someuser", "firstName": "First", "lastName": "Last", "email": None, "roles": []},
            "modifiedBy": {"orcidId": "someuser", "firstName": "First", "lastName": "Last", "email": None, "roles": []},
            "creationDate": date.today().isoformat(),
            "modificationDate": date.today().isoformat(),
            "publishedDate": None,
            "experimentSetUrn": experiment["experimentSetUrn"],
            "doiIdentifiers": [],
            "pubmedIdentifiers": [],
            "rawReadIdentifiers": [],
            "processingState": None,
        },
        "supersededScoreset": None,
        "supersedingScoreset": None,
        "metaAnalysisSourceScoresets": [],
        "metaAnalyses": [],
        "doiIdentifiers": [],
        "pubmedIdentifiers": [],
        "publishedDate": None,
        "creationDate": date.today().isoformat(),
        "modificationDate": date.today().isoformat(),
        "createdBy": {"orcidId": "someuser", "firstName": "First", "lastName": "Last", "email": None, "roles": []},
        "modifiedBy": {"orcidId": "someuser", "firstName": "First", "lastName": "Last", "email": None, "roles": []},
        "targetGene": {
            "name": "UBE2I",
            "category": "Protein coding",
            "externalIdentifiers": [
                {
                    "identifier": {
                        "dbName": "Ensembl",
                        "identifier": "ENSG00000103275",
                        "dbVersion": None,
                        "url": None,
                        "referenceHtml": None,
                    },
                    "offset": 0,
                },
                {
                    "identifier": {
                        "dbName": "RefSeq",
                        "identifier": "NM_003345",
                        "dbVersion": None,
                        "url": None,
                        "referenceHtml": None,
                    },
                    "offset": 159,
                },
                {
                    "identifier": {
                        "dbName": "UniProt",
                        "identifier": "P63279",
                        "dbVersion": None,
                        "url": None,
                        "referenceHtml": None,
                    },
                    "offset": 0,
                },
            ],
            "referenceMaps": [
                {
                    "id": 1,
                    "genomeId": 1,
                    "targetId": 1,
                    "isPrimary": False,
                    "genome": {
                        "shortName": "Name",
                        "organismName": "Organism",
                        "genomeId": None,
                        "creationDate": date.today().isoformat(),
                        "modificationDate": date.today().isoformat(),
                        "id": 1,
                    },
                    "creationDate": date.today().isoformat(),
                    "modificationDate": date.today().isoformat(),
                }
            ],
            "wtSequence": {
                "sequenceType": "dna",
                "sequence": "ATGAGTATTCAACATTTCCGTGTCGCCCTTATTCCCTTTTTTGCGGCATTTTGCCTTCCTGTTTTTGCTCACCCAGAAACGCTGGTGAAAGTAAAAGATGCTGAAGATCAGTTGGGTGCACGAGTGGGTTACATCGAACTGGATCTCAACAGCGGTAAGATCCTTGAGAGTTTTCGCCCCGAAGAACGTTTTCCAATGATGAGCACTTTTAAAGTTCTGCTATGTGGCGCGGTATTATCCCGTGTTGACGCCGGGCAAGAGCAACTCGGTCGCCGCATACACTATTCTCAGAATGACTTGGTTGAGTACTCACCAGTCACAGAAAAGCATCTTACGGATGGCATGACAGTAAGAGAATTATGCAGTGCTGCCATAACCATGAGTGATAACACTGCGGCCAACTTACTTCTGACAACGATCGGAGGACCGAAGGAGCTAACCGCTTTTTTGCACAACATGGGGGATCATGTAACTCGCCTTGATCGTTGGGAACCGGAGCTGAATGAAGCCATACCAAACGACGAGCGTGACACCACGATGCCTGCAGCAATGGCAACAACGTTGCGCAAACTATTAACTGGCGAACTACTTACTCTAGCTTCCCGGCAACAATTAATAGACTGGATGGAGGCGGATAAAGTTGCAGGACCACTTCTGCGCTCGGCCCTTCCGGCTGGCTGGTTTATTGCTGATAAATCTGGAGCCGGTGAGCGTGGGTCTCGCGGTATCATTGCAGCACTGGGGCCAGATGGTAAGCCCTCCCGTATCGTAGTTATCTACACGACGGGGAGTCAGGCAACTATGGATGAACGAAATAGACAGATCGCTGAGATAGGTGCCTCACTGATTAAGCATTGGTAA",
            },
        },
        "datasetColumns": {},
        "private": True,
    }
    assert json.dumps(response_data, sort_keys=True) == json.dumps(expected_response, sort_keys=True)


def test_create_scoreset_with_valid_values(test_empty_db):
    experiment = create_an_experiment()
    create_reference_genome()
    superseded_scoreset = create_scoreset()
    meta_analysis_scoreset = create_scoreset()
    scoreset_to_create = {
        "experimentUrn": experiment["urn"],
        "title": "Test Scoreset Title",
        "methodText": "Methods",
        "abstractText": "Abstract",
        "shortDescription": "Test scoreset",
        "extraMetadata": {"key": "value"},
        "keywords": [],
        "doiIdentifiers": [{"identifier": "10.1016/j.cels.2018.01.015"}, {"identifier": "10.1016/j.cels.2018.01.011"}],
        "pubmedIdentifiers": [{"identifier": "20711194"}, {"identifier": "19502423"}],
        "dataUsagePolicy": "data usage",
        "datasetColumns": {},
        "supersededScoresetUrn": superseded_scoreset["urn"],
        "metaAnalysisSourceScoresetUrns": [meta_analysis_scoreset["urn"]],
        "targetGene": {
            "name": "UBE2I",
            "category": "Protein coding",
            "externalIdentifiers": [
                {"identifier": {"dbName": "Ensembl", "identifier": "ENSG00000103275"}, "offset": 0},
                {"identifier": {"dbName": "RefSeq", "identifier": "NM_003345"}, "offset": 159},
                {"identifier": {"dbName": "UniProt", "identifier": "P63279"}, "offset": 0},
            ],
            "referenceMaps": [{"genomeId": 1}],
            "wtSequence": {
                "sequenceType": "dna",
                "sequence": "ATGAGTATTCAACATTTCCGTGTCGCCCTTATTCCCTTTTTTGCGGCATTTTGCCTTCCTGTTTTTGCTCACCCAGAAACGCTGGTGAAAGTAAAAGATGCTGAAGATCAGTTGGGTGCACGAGTGGGTTACATCGAACTGGATCTCAACAGCGGTAAGATCCTTGAGAGTTTTCGCCCCGAAGAACGTTTTCCAATGATGAGCACTTTTAAAGTTCTGCTATGTGGCGCGGTATTATCCCGTGTTGACGCCGGGCAAGAGCAACTCGGTCGCCGCATACACTATTCTCAGAATGACTTGGTTGAGTACTCACCAGTCACAGAAAAGCATCTTACGGATGGCATGACAGTAAGAGAATTATGCAGTGCTGCCATAACCATGAGTGATAACACTGCGGCCAACTTACTTCTGACAACGATCGGAGGACCGAAGGAGCTAACCGCTTTTTTGCACAACATGGGGGATCATGTAACTCGCCTTGATCGTTGGGAACCGGAGCTGAATGAAGCCATACCAAACGACGAGCGTGACACCACGATGCCTGCAGCAATGGCAACAACGTTGCGCAAACTATTAACTGGCGAACTACTTACTCTAGCTTCCCGGCAACAATTAATAGACTGGATGGAGGCGGATAAAGTTGCAGGACCACTTCTGCGCTCGGCCCTTCCGGCTGGCTGGTTTATTGCTGATAAATCTGGAGCCGGTGAGCGTGGGTCTCGCGGTATCATTGCAGCACTGGGGCCAGATGGTAAGCCCTCCCGTATCGTAGTTATCTACACGACGGGGAGTCAGGCAACTATGGATGAACGAAATAGACAGATCGCTGAGATAGGTGCCTCACTGATTAAGCATTGGTAA",
            },
        },
    }
    response = client.post("/api/v1/scoresets/", json=scoreset_to_create)
    assert response.status_code == 200
    response_data = response.json()
    scoreset_urn = response_data["urn"]
    experiment_urn = experiment["urn"]
    assert scoreset_urn is not None
    assert experiment_urn is not None
    assert re.match(r"tmp:[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}", scoreset_urn)
    assert re.match(r"tmp:[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}", experiment_urn)
    expected_response = {
        "title": "Test Scoreset Title",
        "methodText": "Methods",
        "abstractText": "Abstract",
        "shortDescription": "Test scoreset",
        "extraMetadata": {"key": "value"},
        "dataUsagePolicy": "data usage",
        "licenceId": None,
        "keywords": [],
        "urn": scoreset_urn,
        "numVariants": 0,
        "experiment": {
            "title": "Test Experiment Title",
            "shortDescription": "Test experiment",
            "abstractText": "Abstract",
            "methodText": "Methods",
            "extraMetadata": {"key": "value"},
            "keywords": [],
            "urn": experiment_urn,
            "numScoresets": 0,
            "createdBy": {"orcidId": "someuser", "firstName": "First", "lastName": "Last", "email": None, "roles": []},
            "modifiedBy": {"orcidId": "someuser", "firstName": "First", "lastName": "Last", "email": None, "roles": []},
            "creationDate": date.today().isoformat(),
            "modificationDate": date.today().isoformat(),
            "publishedDate": None,
            "experimentSetUrn": experiment["experimentSetUrn"],
            "doiIdentifiers": [],
            "pubmedIdentifiers": [],
            "rawReadIdentifiers": [],
            "processingState": None,
        },
        "supersededScoreset": {
            "urn": superseded_scoreset["urn"],
            "title": "Test Scoreset Title",
            "shortDescription": "Test scoreset",
            "publishedDate": None,
            "replacesId": None,
            "numVariants": 0,
            "experiment": {
                "title": "Test Experiment Title",
                "shortDescription": "Test experiment",
                "abstractText": "Abstract",
                "methodText": "Methods",
                "extraMetadata": {"key": "value"},
                "keywords": [],
                "urn": superseded_scoreset["experiment"]["urn"],
                "numScoresets": 0,
                "createdBy": {
                    "orcidId": "someuser",
                    "firstName": "First",
                    "lastName": "Last",
                    "email": None,
                    "roles": [],
                },
                "modifiedBy": {
                    "orcidId": "someuser",
                    "firstName": "First",
                    "lastName": "Last",
                    "email": None,
                    "roles": [],
                },
                "creationDate": date.today().isoformat(),
                "modificationDate": date.today().isoformat(),
                "publishedDate": None,
                "experimentSetUrn": superseded_scoreset["experiment"]["experimentSetUrn"],
                "doiIdentifiers": [],
                "pubmedIdentifiers": [],
                "rawReadIdentifiers": [],
                "processingState": None,
            },
            "creationDate": date.today().isoformat(),
            "modificationDate": date.today().isoformat(),
            "targetGene": {
                "name": "UBE2I",
                "category": "Protein coding",
                "externalIdentifiers": [
                    {
                        "identifier": {
                            "dbName": "Ensembl",
                            "identifier": "ENSG00000103275",
                            "dbVersion": None,
                            "url": None,
                            "referenceHtml": None,
                        },
                        "offset": 0,
                    },
                    {
                        "identifier": {
                            "dbName": "RefSeq",
                            "identifier": "NM_003345",
                            "dbVersion": None,
                            "url": None,
                            "referenceHtml": None,
                        },
                        "offset": 159,
                    },
                    {
                        "identifier": {
                            "dbName": "UniProt",
                            "identifier": "P63279",
                            "dbVersion": None,
                            "url": None,
                            "referenceHtml": None,
                        },
                        "offset": 0,
                    },
                ],
                "referenceMaps": [
                    {
                        "id": 1,
                        "genomeId": 1,
                        "targetId": 1,
                        "isPrimary": False,
                        "genome": {
                            "shortName": "Name",
                            "organismName": "Organism",
                            "genomeId": None,
                            "creationDate": date.today().isoformat(),
                            "modificationDate": date.today().isoformat(),
                            "id": 1,
                        },
                        "creationDate": date.today().isoformat(),
                        "modificationDate": date.today().isoformat(),
                    }
                ],
            },
            "private": True,
        },
        "supersedingScoreset": None,
        "metaAnalysisSourceScoresets": [
            {
                "urn": meta_analysis_scoreset["urn"],
                "title": "Test Scoreset Title",
                "shortDescription": "Test scoreset",
                "publishedDate": None,
                "replacesId": None,
                "numVariants": 0,
                "experiment": {
                    "title": "Test Experiment Title",
                    "shortDescription": "Test experiment",
                    "abstractText": "Abstract",
                    "methodText": "Methods",
                    "extraMetadata": {"key": "value"},
                    "keywords": [],
                    "urn": meta_analysis_scoreset["experiment"]["urn"],
                    "numScoresets": 0,
                    "createdBy": {
                        "orcidId": "someuser",
                        "firstName": "First",
                        "lastName": "Last",
                        "email": None,
                        "roles": [],
                    },
                    "modifiedBy": {
                        "orcidId": "someuser",
                        "firstName": "First",
                        "lastName": "Last",
                        "email": None,
                        "roles": [],
                    },
                    "creationDate": date.today().isoformat(),
                    "modificationDate": date.today().isoformat(),
                    "publishedDate": None,
                    "experimentSetUrn": meta_analysis_scoreset["experiment"]["experimentSetUrn"],
                    "doiIdentifiers": [],
                    "pubmedIdentifiers": [],
                    "rawReadIdentifiers": [],
                    "processingState": None,
                },
                "creationDate": date.today().isoformat(),
                "modificationDate": date.today().isoformat(),
                "targetGene": {
                    "name": "UBE2I",
                    "category": "Protein coding",
                    "externalIdentifiers": [
                        {
                            "identifier": {
                                "dbName": "Ensembl",
                                "identifier": "ENSG00000103275",
                                "dbVersion": None,
                                "url": None,
                                "referenceHtml": None,
                            },
                            "offset": 0,
                        },
                        {
                            "identifier": {
                                "dbName": "RefSeq",
                                "identifier": "NM_003345",
                                "dbVersion": None,
                                "url": None,
                                "referenceHtml": None,
                            },
                            "offset": 159,
                        },
                        {
                            "identifier": {
                                "dbName": "UniProt",
                                "identifier": "P63279",
                                "dbVersion": None,
                                "url": None,
                                "referenceHtml": None,
                            },
                            "offset": 0,
                        },
                    ],
                    "referenceMaps": [
                        {
                            "id": 2,
                            "genomeId": 1,
                            "targetId": 2,
                            "isPrimary": False,
                            "genome": {
                                "shortName": "Name",
                                "organismName": "Organism",
                                "genomeId": None,
                                "creationDate": date.today().isoformat(),
                                "modificationDate": date.today().isoformat(),
                                "id": 1,
                            },
                            "creationDate": date.today().isoformat(),
                            "modificationDate": date.today().isoformat(),
                        }
                    ],
                },
                "private": True,
            }
        ],
        "metaAnalyses": [],
        "doiIdentifiers": [
            {"identifier": "10.1016/j.cels.2018.01.015", "id": 1, "url": "https://doi.org/10.1016/j.cels.2018.01.015"},
            {"identifier": "10.1016/j.cels.2018.01.011", "id": 2, "url": "https://doi.org/10.1016/j.cels.2018.01.011"},
        ],
        "pubmedIdentifiers": [
            {
                "identifier": "20711194",
                "id": 1,
                "url": "http://www.ncbi.nlm.nih.gov/pubmed/20711194",
                "referenceHtml": "Fowler DM, <i>et al</i>. High-resolution mapping of protein sequence-function relationships. High-resolution mapping of protein sequence-function relationships. 2010; 7:741-6. doi: 10.1038/nmeth.1492",
            },
            {
                "identifier": "19502423",
                "id": 2,
                "url": "http://www.ncbi.nlm.nih.gov/pubmed/19502423",
                "referenceHtml": "Sohka T, <i>et al</i>. An externally tunable bacterial band-pass filter. An externally tunable bacterial band-pass filter. 2009; 106:10135-40. doi: 10.1073/pnas.0901246106",
            },
        ],
        "publishedDate": None,
        "creationDate": date.today().isoformat(),
        "modificationDate": date.today().isoformat(),
        "createdBy": {"orcidId": "someuser", "firstName": "First", "lastName": "Last", "email": None, "roles": []},
        "modifiedBy": {"orcidId": "someuser", "firstName": "First", "lastName": "Last", "email": None, "roles": []},
        "targetGene": {
            "name": "UBE2I",
            "category": "Protein coding",
            "externalIdentifiers": [
                {
                    "identifier": {
                        "dbName": "Ensembl",
                        "identifier": "ENSG00000103275",
                        "dbVersion": None,
                        "url": None,
                        "referenceHtml": None,
                    },
                    "offset": 0,
                },
                {
                    "identifier": {
                        "dbName": "RefSeq",
                        "identifier": "NM_003345",
                        "dbVersion": None,
                        "url": None,
                        "referenceHtml": None,
                    },
                    "offset": 159,
                },
                {
                    "identifier": {
                        "dbName": "UniProt",
                        "identifier": "P63279",
                        "dbVersion": None,
                        "url": None,
                        "referenceHtml": None,
                    },
                    "offset": 0,
                },
            ],
            "referenceMaps": [
                {
                    "id": 3,
                    "genomeId": 1,
                    "targetId": 3,
                    "isPrimary": False,
                    "genome": {
                        "shortName": "Name",
                        "organismName": "Organism",
                        "genomeId": None,
                        "creationDate": date.today().isoformat(),
                        "modificationDate": date.today().isoformat(),
                        "id": 1,
                    },
                    "creationDate": date.today().isoformat(),
                    "modificationDate": date.today().isoformat(),
                }
            ],
            "wtSequence": {
                "sequenceType": "dna",
                "sequence": "ATGAGTATTCAACATTTCCGTGTCGCCCTTATTCCCTTTTTTGCGGCATTTTGCCTTCCTGTTTTTGCTCACCCAGAAACGCTGGTGAAAGTAAAAGATGCTGAAGATCAGTTGGGTGCACGAGTGGGTTACATCGAACTGGATCTCAACAGCGGTAAGATCCTTGAGAGTTTTCGCCCCGAAGAACGTTTTCCAATGATGAGCACTTTTAAAGTTCTGCTATGTGGCGCGGTATTATCCCGTGTTGACGCCGGGCAAGAGCAACTCGGTCGCCGCATACACTATTCTCAGAATGACTTGGTTGAGTACTCACCAGTCACAGAAAAGCATCTTACGGATGGCATGACAGTAAGAGAATTATGCAGTGCTGCCATAACCATGAGTGATAACACTGCGGCCAACTTACTTCTGACAACGATCGGAGGACCGAAGGAGCTAACCGCTTTTTTGCACAACATGGGGGATCATGTAACTCGCCTTGATCGTTGGGAACCGGAGCTGAATGAAGCCATACCAAACGACGAGCGTGACACCACGATGCCTGCAGCAATGGCAACAACGTTGCGCAAACTATTAACTGGCGAACTACTTACTCTAGCTTCCCGGCAACAATTAATAGACTGGATGGAGGCGGATAAAGTTGCAGGACCACTTCTGCGCTCGGCCCTTCCGGCTGGCTGGTTTATTGCTGATAAATCTGGAGCCGGTGAGCGTGGGTCTCGCGGTATCATTGCAGCACTGGGGCCAGATGGTAAGCCCTCCCGTATCGTAGTTATCTACACGACGGGGAGTCAGGCAACTATGGATGAACGAAATAGACAGATCGCTGAGATAGGTGCCTCACTGATTAAGCATTGGTAA",
            },
        },
        "datasetColumns": {},
        "private": True,
    }
    assert json.dumps(response_data, sort_keys=True) == json.dumps(expected_response, sort_keys=True)


def test_create_scoreset_with_invalid_doi(test_empty_db):
    experiment = create_an_experiment()
    create_reference_genome()
    scoreset_to_create = {
        "experimentUrn": experiment["urn"],
        "title": "Test Scoreset Title",
        "methodText": "Methods",
        "abstractText": "Abstract",
        "shortDescription": "Test scoreset",
        "extraMetadata": {"key": "value"},
        "keywords": [],
        "doiIdentifiers": [{"identifier": "abab"}],
        "pubmedIdentifiers": [],
        "dataUsagePolicy": None,
        "datasetColumns": {},
        "supersededScoresetUrn": None,
        "metaAnalysisSourceScoresetUrns": [],
        "targetGene": {
            "name": "UBE2I",
            "category": "Protein coding",
            "externalIdentifiers": [
                {"identifier": {"dbName": "Ensembl", "identifier": "ENSG00000103275"}, "offset": 0},
                {"identifier": {"dbName": "RefSeq", "identifier": "NM_003345"}, "offset": 159},
                {"identifier": {"dbName": "UniProt", "identifier": "P63279"}, "offset": 0},
            ],
            "referenceMaps": [{"genomeId": 1}],
            "wtSequence": {
                "sequenceType": "dna",
                "sequence": "ATGAGTATTCAACATTTCCGTGTCGCCCTTATTCCCTTTTTTGCGGCATTTTGCCTTCCTGTTTTTGCTCACCCAGAAACGCTGGTGAAAGTAAAAGATGCTGAAGATCAGTTGGGTGCACGAGTGGGTTACATCGAACTGGATCTCAACAGCGGTAAGATCCTTGAGAGTTTTCGCCCCGAAGAACGTTTTCCAATGATGAGCACTTTTAAAGTTCTGCTATGTGGCGCGGTATTATCCCGTGTTGACGCCGGGCAAGAGCAACTCGGTCGCCGCATACACTATTCTCAGAATGACTTGGTTGAGTACTCACCAGTCACAGAAAAGCATCTTACGGATGGCATGACAGTAAGAGAATTATGCAGTGCTGCCATAACCATGAGTGATAACACTGCGGCCAACTTACTTCTGACAACGATCGGAGGACCGAAGGAGCTAACCGCTTTTTTGCACAACATGGGGGATCATGTAACTCGCCTTGATCGTTGGGAACCGGAGCTGAATGAAGCCATACCAAACGACGAGCGTGACACCACGATGCCTGCAGCAATGGCAACAACGTTGCGCAAACTATTAACTGGCGAACTACTTACTCTAGCTTCCCGGCAACAATTAATAGACTGGATGGAGGCGGATAAAGTTGCAGGACCACTTCTGCGCTCGGCCCTTCCGGCTGGCTGGTTTATTGCTGATAAATCTGGAGCCGGTGAGCGTGGGTCTCGCGGTATCATTGCAGCACTGGGGCCAGATGGTAAGCCCTCCCGTATCGTAGTTATCTACACGACGGGGAGTCAGGCAACTATGGATGAACGAAATAGACAGATCGCTGAGATAGGTGCCTCACTGATTAAGCATTGGTAA",
            },
        },
    }
    response = client.post("/api/v1/scoresets/", json=scoreset_to_create)
    assert response.status_code == 422


def test_create_scoreset_with_invalid_pubmed(test_empty_db):
    experiment = create_an_experiment()
    create_reference_genome()
    scoreset_to_create = {
        "experimentUrn": experiment["urn"],
        "title": "Test Scoreset Title",
        "methodText": "Methods",
        "abstractText": "Abstract",
        "shortDescription": "Test scoreset",
        "extraMetadata": {"key": "value"},
        "keywords": [],
        "doiIdentifiers": [],
        "pubmedIdentifiers": [{"identifier": "abcc"}],
        "dataUsagePolicy": None,
        "datasetColumns": {},
        "supersededScoresetUrn": None,
        "metaAnalysisSourceScoresetUrns": [],
        "targetGene": {
            "name": "UBE2I",
            "category": "Protein coding",
            "externalIdentifiers": [
                {"identifier": {"dbName": "Ensembl", "identifier": "ENSG00000103275"}, "offset": 0},
                {"identifier": {"dbName": "RefSeq", "identifier": "NM_003345"}, "offset": 159},
                {"identifier": {"dbName": "UniProt", "identifier": "P63279"}, "offset": 0},
            ],
            "referenceMaps": [{"genomeId": 1}],
            "wtSequence": {
                "sequenceType": "dna",
                "sequence": "ATGAGTATTCAACATTTCCGTGTCGCCCTTATTCCCTTTTTTGCGGCATTTTGCCTTCCTGTTTTTGCTCACCCAGAAACGCTGGTGAAAGTAAAAGATGCTGAAGATCAGTTGGGTGCACGAGTGGGTTACATCGAACTGGATCTCAACAGCGGTAAGATCCTTGAGAGTTTTCGCCCCGAAGAACGTTTTCCAATGATGAGCACTTTTAAAGTTCTGCTATGTGGCGCGGTATTATCCCGTGTTGACGCCGGGCAAGAGCAACTCGGTCGCCGCATACACTATTCTCAGAATGACTTGGTTGAGTACTCACCAGTCACAGAAAAGCATCTTACGGATGGCATGACAGTAAGAGAATTATGCAGTGCTGCCATAACCATGAGTGATAACACTGCGGCCAACTTACTTCTGACAACGATCGGAGGACCGAAGGAGCTAACCGCTTTTTTGCACAACATGGGGGATCATGTAACTCGCCTTGATCGTTGGGAACCGGAGCTGAATGAAGCCATACCAAACGACGAGCGTGACACCACGATGCCTGCAGCAATGGCAACAACGTTGCGCAAACTATTAACTGGCGAACTACTTACTCTAGCTTCCCGGCAACAATTAATAGACTGGATGGAGGCGGATAAAGTTGCAGGACCACTTCTGCGCTCGGCCCTTCCGGCTGGCTGGTTTATTGCTGATAAATCTGGAGCCGGTGAGCGTGGGTCTCGCGGTATCATTGCAGCACTGGGGCCAGATGGTAAGCCCTCCCGTATCGTAGTTATCTACACGACGGGGAGTCAGGCAACTATGGATGAACGAAATAGACAGATCGCTGAGATAGGTGCCTCACTGATTAAGCATTGGTAA",
            },
        },
    }
    response = client.post("/api/v1/scoresets/", json=scoreset_to_create)
    assert response.status_code == 422


def test_create_scoreset_with_invalid_experiment(test_empty_db):
    create_reference_genome()
    scoreset_to_create = {
        "experimentUrn": "invalid_urn",
        "title": "Test Scoreset Title",
        "methodText": "Methods",
        "abstractText": "Abstract",
        "shortDescription": "Test scoreset",
        "extraMetadata": {"key": "value"},
        "keywords": [],
        "doiIdentifiers": [],
        "pubmedIdentifiers": [],
        "dataUsagePolicy": None,
        "datasetColumns": {},
        "supersededScoresetUrn": None,
        "metaAnalysisSourceScoresetUrns": [],
        "targetGene": {
            "name": "UBE2I",
            "category": "Protein coding",
            "externalIdentifiers": [
                {"identifier": {"dbName": "Ensembl", "identifier": "ENSG00000103275"}, "offset": 0},
                {"identifier": {"dbName": "RefSeq", "identifier": "NM_003345"}, "offset": 159},
                {"identifier": {"dbName": "UniProt", "identifier": "P63279"}, "offset": 0},
            ],
            "referenceMaps": [{"genomeId": 1}],
            "wtSequence": {
                "sequenceType": "dna",
                "sequence": "ATGAGTATTCAACATTTCCGTGTCGCCCTTATTCCCTTTTTTGCGGCATTTTGCCTTCCTGTTTTTGCTCACCCAGAAACGCTGGTGAAAGTAAAAGATGCTGAAGATCAGTTGGGTGCACGAGTGGGTTACATCGAACTGGATCTCAACAGCGGTAAGATCCTTGAGAGTTTTCGCCCCGAAGAACGTTTTCCAATGATGAGCACTTTTAAAGTTCTGCTATGTGGCGCGGTATTATCCCGTGTTGACGCCGGGCAAGAGCAACTCGGTCGCCGCATACACTATTCTCAGAATGACTTGGTTGAGTACTCACCAGTCACAGAAAAGCATCTTACGGATGGCATGACAGTAAGAGAATTATGCAGTGCTGCCATAACCATGAGTGATAACACTGCGGCCAACTTACTTCTGACAACGATCGGAGGACCGAAGGAGCTAACCGCTTTTTTGCACAACATGGGGGATCATGTAACTCGCCTTGATCGTTGGGAACCGGAGCTGAATGAAGCCATACCAAACGACGAGCGTGACACCACGATGCCTGCAGCAATGGCAACAACGTTGCGCAAACTATTAACTGGCGAACTACTTACTCTAGCTTCCCGGCAACAATTAATAGACTGGATGGAGGCGGATAAAGTTGCAGGACCACTTCTGCGCTCGGCCCTTCCGGCTGGCTGGTTTATTGCTGATAAATCTGGAGCCGGTGAGCGTGGGTCTCGCGGTATCATTGCAGCACTGGGGCCAGATGGTAAGCCCTCCCGTATCGTAGTTATCTACACGACGGGGAGTCAGGCAACTATGGATGAACGAAATAGACAGATCGCTGAGATAGGTGCCTCACTGATTAAGCATTGGTAA",
            },
        },
    }
    response = client.post("/api/v1/scoresets/", json=scoreset_to_create)
    assert response.status_code == 422


def test_show_scoreset(test_empty_db):
    create_reference_genome()
    scoreset = create_scoreset()
    urn = scoreset["urn"]
    response = client.get(f"/api/v1/scoresets/{urn}")
    assert response.status_code == 200


# Can't pass
def test_valid_score_file(test_empty_db):
    create_reference_genome()
    scoreset = create_scoreset()
    urn = scoreset["urn"]
    current_directory = os.path.dirname(os.path.abspath(__file__))
    score_csv_path = os.path.join(current_directory, "scores.csv")
    count_csv_path = os.path.join(current_directory, "counts.csv")
    # csv_path = os.path.join("/Users/da.e/Desktop/Estelle Computer/Programming Work/mavedb-api/app/tests/", "scores.csv")
    with open(score_csv_path, "rb") as f1, open(count_csv_path, "rb") as f2:
        response = client.post(
            f"/api/v1/scoresets/{urn}/variants/data",
            files={"scores_file": ("scores.csv", f1, "text/csv"), "counts_file": ("counts.csv", f2, "text/csv")},
        )
    # check the response status code and data
    print(response.json())
    assert response.status_code == 200


"""
# This one can't work cause duplicate column names will be processed automatically
def test_score_file_with_duplicate_columns(test_empty_db):
    create_reference_genome()
    scoreset = create_scoreset()
    urn = scoreset["urn"]
    current_directory = os.path.dirname(os.path.abspath(__file__))
    score_csv_path = os.path.join(current_directory, "scores_with_duplicate_columns.csv")
    count_csv_path = os.path.join(current_directory, "counts.csv")
    # csv_path = os.path.join("/Users/da.e/Desktop/Estelle Computer/Programming Work/mavedb-api/app/tests/", "scores.csv")
    with open(score_csv_path, "rb") as f1, open(count_csv_path, "rb") as f2:
        response = client.post(f"/api/v1/scoresets/{urn}/variants/data", files={'scores_file': ('scores.csv', f1, "text/csv"),
                                                                                'counts_file': ('counts.csv', f2, "text/csv")})
    # check the response status code and data
    print(response.json())
    assert response.status_code == 400
    assert response.json() == {"detail": "There cannot be duplicate column names."}
"""


# Is it a ccorect way?
def test_score_file_without_score_column(test_empty_db):
    create_reference_genome()
    scoreset = create_scoreset()
    urn = scoreset["urn"]
    current_directory = os.path.dirname(os.path.abspath(__file__))
    score_csv_path = os.path.join(current_directory, "scores_without_score_column.csv")
    with open(score_csv_path, "rb") as f:
        response = client.post(
            f"/api/v1/scoresets/{urn}/variants/data",
            files={"scores_file": ("scores_without_score_column.csv", f, "text/csv")},
        )
    assert response.status_code == 400
    assert response.json() == {"detail": "score dataframe must have a 'score' column"}


# Show other problem. Can't pass.
def test_scores_and_counts_define_different_variants(test_empty_db):
    create_reference_genome()
    scoreset = create_scoreset()
    urn = scoreset["urn"]
    current_directory = os.path.dirname(os.path.abspath(__file__))
    score_csv_path = os.path.join(current_directory, "scores.csv")
    count_csv_path = os.path.join(current_directory, "counts_with_different_variants.csv")
    with open(score_csv_path, "rb") as f1, open(count_csv_path, "rb") as f2:
        response = client.post(
            f"/api/v1/scoresets/{urn}/variants/data",
            files={
                "scores_file": ("scores.csv", f1, "text/csv"),
                "counts_file": ("counts_with_different_variants.csv", f2, "text/csv"),
            },
        )
    assert response.status_code == 400
    assert response.json() == {
        "detail": "both score and count dataframes must define matching variants, discrepancy found in 'hgvs_pro'"
    }


def test_score_file_with_letter(test_empty_db):
    create_reference_genome()
    scoreset = create_scoreset()
    urn = scoreset["urn"]
    current_directory = os.path.dirname(os.path.abspath(__file__))
    score_csv_path = os.path.join(current_directory, "scores_with_string.csv")
    with open(score_csv_path, "rb") as f:
        response = client.post(
            f"/api/v1/scoresets/{urn}/variants/data", files={"scores_file": ("scores_with_string.csv", f, "text/csv")}
        )
    assert response.status_code == 400
    assert response.json() == {"detail": "data column 'score' has mixed string and numeric types"}


# can't catch error
def test_score_file_without_hgvs_columns(test_empty_db):
    create_reference_genome()
    scoreset = create_scoreset()
    urn = scoreset["urn"]
    current_directory = os.path.dirname(os.path.abspath(__file__))
    score_csv_path = os.path.join(current_directory, "scores_without_hgvs_column.csv")
    with open(score_csv_path, "rb") as f:
        response = client.post(
            f"/api/v1/scoresets/{urn}/variants/data",
            files={"scores_file": ("scores_without_hgvs_column.csv", f, "text/csv")},
        )
    assert response.status_code == 400
    # I'm confused. There're two validations about this. This one is in validate_column_names
    # The other one is in validate_values_by_column. It raises ValidationError("Missing required hgvs and/or score columns.")
    # And then there's one more ValidationError("Must include either hgvs_nt or hgvs_pro column.")
    assert response.json() == {"detail": "failed to find valid HGVS variant column"}


def test_score_file_hgvs_pro_has_same_values(test_empty_db):
    create_reference_genome()
    scoreset = create_scoreset()
    urn = scoreset["urn"]
    current_directory = os.path.dirname(os.path.abspath(__file__))
    score_csv_path = os.path.join(current_directory, "scores_hgvs_pro_has_same_values.csv")
    with open(score_csv_path, "rb") as f:
        response = client.post(
            f"/api/v1/scoresets/{urn}/variants/data",
            files={"scores_file": ("scores_hgvs_pro_has_same_values.csv", f, "text/csv")},
        )
    assert response.status_code == 400
    assert response.json() == {"detail": "primary variant column 'hgvs_pro' must contain unique values"}


def test_score_and_count_files_have_hgvs_nt_and_pro(test_empty_db):
    create_reference_genome()
    scoreset = create_scoreset()
    urn = scoreset["urn"]
    current_directory = os.path.dirname(os.path.abspath(__file__))
    score_csv_path = os.path.join(current_directory, "scores_with_hgvs_nt_and_pro.csv")
    count_csv_path = os.path.join(current_directory, "counts_with_hgvs_nt_and_pro.csv")
    with open(score_csv_path, "rb") as f1, open(count_csv_path, "rb") as f2:
        response = client.post(
            f"/api/v1/scoresets/{urn}/variants/data",
            files={
                "scores_file": ("scores_with_hgvs_nt_and_pro.csv", f1, "text/csv"),
                "counts_file": ("counts_with_hgvs_nt_and_pro.csv", f2, "text/csv"),
            },
        )
    assert response.status_code == 200


# Can't catch error.
def test_score_file_has_not_match_hgvs_nt_and_pro(test_empty_db):
    create_reference_genome()
    scoreset = create_scoreset()
    urn = scoreset["urn"]
    current_directory = os.path.dirname(os.path.abspath(__file__))
    score_csv_path = os.path.join(current_directory, "scores_hgvs_nt_not_match_pro.csv")
    with open(score_csv_path, "rb") as f:
        response = client.post(
            f"/api/v1/scoresets/{urn}/variants/data",
            files={"scores_file": ("scores_hgvs_nt_not_match_pro.csv", f, "text/csv")},
        )
    assert response.status_code == 400
    # assert response.json() == {"detail": "Each value in hgvs_pro column must be unique."}


def test_score_file_has_invalid_hgvs_pro_prefix(test_empty_db):
    create_reference_genome()
    scoreset = create_scoreset()
    urn = scoreset["urn"]
    current_directory = os.path.dirname(os.path.abspath(__file__))
    score_csv_path = os.path.join(current_directory, "scores_with_invalid_hgvs_pro_prefix.csv")
    with open(score_csv_path, "rb") as f:
        response = client.post(
            f"/api/v1/scoresets/{urn}/variants/data",
            files={"scores_file": ("scores_with_invalid_hgvs_pro_prefix.csv", f, "text/csv")},
        )
    assert response.status_code == 400


def test_score_file_has_invalid_hgvs_nt_prefix(test_empty_db):
    create_reference_genome()
    scoreset = create_scoreset()
    urn = scoreset["urn"]
    current_directory = os.path.dirname(os.path.abspath(__file__))
    score_csv_path = os.path.join(current_directory, "scores_with_invalid_hgvs_nt_prefix.csv")
    with open(score_csv_path, "rb") as f:
        response = client.post(
            f"/api/v1/scoresets/{urn}/variants/data",
            files={"scores_file": ("scores_with_invalid_hgvs_nt_prefix.csv", f, "text/csv")},
        )
    assert response.status_code == 400


def test_count_file_has_score_column(test_empty_db):
    create_reference_genome()
    scoreset = create_scoreset()
    urn = scoreset["urn"]
    current_directory = os.path.dirname(os.path.abspath(__file__))
    score_csv_path = os.path.join(current_directory, "scores.csv")
    count_csv_path = os.path.join(current_directory, "counts_with_score.csv")
    with open(score_csv_path, "rb") as f1, open(count_csv_path, "rb") as f2:
        response = client.post(
            f"/api/v1/scoresets/{urn}/variants/data",
            files={
                "scores_file": ("scores.csv", f1, "text/csv"),
                "counts_file": ("counts_with_score.csv", f2, "text/csv"),
            },
        )
    assert response.status_code == 400
    assert response.json() == {"detail": "counts dataframe must not have a 'score' column"}


def test_score_file_column_name_has_nan(test_empty_db):
    create_reference_genome()
    scoreset = create_scoreset()
    urn = scoreset["urn"]
    current_directory = os.path.dirname(os.path.abspath(__file__))
    score_csv_path = os.path.join(current_directory, "scores_with_nan_column_name.csv")
    with open(score_csv_path, "rb") as f:
        response = client.post(
            f"/api/v1/scoresets/{urn}/variants/data",
            files={"scores_file": ("scores_with_nan_column_name.csv", f, "text/csv")},
        )
    assert response.status_code == 400
    assert response.json() == {"detail": "column names cannot be empty or whitespace"}



from typing import Any
from operator import attrgetter

from fastapi import APIRouter, Depends, HTTPException
from sqlalchemy.orm import Session

from mavedb import deps
from mavedb.models.experiment_set import ExperimentSet
from mavedb.view_models import experiment_set

router = APIRouter(
    prefix="/api/v1/experimentSets", tags=["experiment-sets"], responses={404: {"description": "Not found"}}
)


@router.get("/{urn}", status_code=200, response_model=experiment_set.ExperimentSet, responses={404: {}})
def fetch_experiment_set(*, urn: str, db: Session = Depends(deps.get_db)) -> Any:
    """
    Fetch a single experiment set by URN.
    """
    # result = crud.scoreset.get(db=db, id=id)
    # item = db.query(ExperimentSet).filter(ExperimentSet.urn == urn).filter(ExperimentSet.private.is_(False)).first()
    item = db.query(ExperimentSet).filter(ExperimentSet.urn == urn).first()
    if not item:
        # the exception is raised, not returned - you will get a validation
        # error otherwise.
        raise HTTPException(status_code=404, detail=f"Experiment set with URN {urn} not found")
    else:
        item.experiments.sort(key=attrgetter("urn"))
    return item



from datetime import date

from sqlalchemy import Boolean, Column, Date, ForeignKey, Integer, String
from sqlalchemy.orm import relationship
from sqlalchemy.schema import Table

from mavedb.db.base import Base
from mavedb.deps import JSONB
from mavedb.lib.temp_urns import generate_temp_urn

experiment_sets_doi_identifiers_association_table = Table(
    "experiment_set_doi_identifiers",
    Base.metadata,
    Column("experiment_set_id", ForeignKey("experiment_sets.id"), primary_key=True),
    Column("doi_identifier_id", ForeignKey("doi_identifiers.id"), primary_key=True),
)

experiment_sets_keywords_association_table = Table(
    "experiment_set_keywords",
    Base.metadata,
    Column("experiment_set_id", ForeignKey("experiment_sets.id"), primary_key=True),
    Column("keyword_id", ForeignKey("keywords.id"), primary_key=True),
)

experiment_sets_pubmed_identifiers_association_table = Table(
    "experiment_set_pubmed_identifiers",
    Base.metadata,
    Column("experiment_set_id", ForeignKey("experiment_sets.id"), primary_key=True),
    Column("pubmed_identifier_id", ForeignKey("pubmed_identifiers.id"), primary_key=True),
)

# experiment_sets_sra_identifiers_association_table = Table(
experiment_sets_raw_read_identifiers_association_table = Table(
    "experiment_set_sra_identifiers",
    Base.metadata,
    Column("experiment_set_id", ForeignKey("experiment_sets.id"), primary_key=True),
    Column("sra_identifier_id", ForeignKey("sra_identifiers.id"), primary_key=True),
)


class ExperimentSet(Base):
    __tablename__ = "experiment_sets"

    id = Column(Integer, primary_key=True, index=True)

    urn = Column(String(64), nullable=True, default=generate_temp_urn)  # index=True, nullable=True
    title = Column(String, nullable=False)
    method_text = Column(String, nullable=False)
    abstract_text = Column(String, nullable=False)
    short_description = Column(String, nullable=False)
    extra_metadata = Column(JSONB, nullable=False)

    private = Column(Boolean, nullable=False, default=True)
    approved = Column(Boolean, nullable=False, default=False)
    published_date = Column(Date, nullable=True)
    processing_state = Column(String, nullable=True)

    # TODO Refactor the way we handle child collections?
    num_experiments = Column(Integer, nullable=False, default=0)

    created_by_id = Column(Integer, ForeignKey("users.id"), nullable=True)
    created_by = relationship("User", foreign_keys="ExperimentSet.created_by_id")
    modified_by_id = Column(Integer, ForeignKey("users.id"), nullable=True)
    modified_by = relationship("User", foreign_keys="ExperimentSet.modified_by_id")
    creation_date = Column(Date, nullable=False, default=date.today)
    modification_date = Column(Date, nullable=False, default=date.today, onupdate=date.today)

    keyword_objs = relationship(
        "Keyword", secondary=experiment_sets_keywords_association_table, backref="experiment_sets"
    )
    doi_identifiers = relationship(
        "DoiIdentifier", secondary=experiment_sets_doi_identifiers_association_table, backref="experiment_sets"
    )
    pubmed_identifiers = relationship(
        "PubmedIdentifier", secondary=experiment_sets_pubmed_identifiers_association_table, backref="experiment_sets"
    )
    # sra_identifiers = relationship('SraIdentifier', secondary=experiment_sets_sra_identifiers_association_table, backref='experiment_sets')
    raw_read_identifiers = relationship(
        "RawReadIdentifier", secondary=experiment_sets_raw_read_identifiers_association_table, backref="experiment_sets"
    )



from mavedb.view_models.base.base import BaseModel


class RawReadIdentifierBase(BaseModel):
    identifier: str


class RawReadIdentifierCreate(RawReadIdentifierBase):
    pass


# Properties shared by models stored in DB
class SavedRawReadIdentifier(RawReadIdentifierBase):
    id: int
    url: str

    class Config:
        orm_mode = True


# Properties to return to non-admin clients
class RawReadIdentifier(SavedRawReadIdentifier):
    pass



from typing import Any, List

from fastapi import APIRouter, Depends, HTTPException
from sqlalchemy import func
from sqlalchemy.orm import Session

from mavedb import deps
from mavedb.models.doi_identifier import DoiIdentifier
from mavedb.view_models import doi_identifier
from mavedb.view_models.search import TextSearch

router = APIRouter(
    prefix="/api/v1/doiIdentifiers", tags=["DOI identifiers"], responses={404: {"description": "Not found"}}
)


@router.post("/search", status_code=200, response_model=List[doi_identifier.DoiIdentifier])
def search_doi_identifiers(search: TextSearch, db: Session = Depends(deps.get_db)) -> Any:
    """
    Search DOI identifiers.
    """

    query = db.query(DoiIdentifier)

    if search.text and len(search.text.strip()) > 0:
        lower_search_text = search.text.strip().lower()
        query = query.filter(func.lower(DoiIdentifier.identifier).contains(lower_search_text))
    else:
        raise HTTPException(status_code=500, detail="Search text is required")

    items = query.order_by(DoiIdentifier.identifier).limit(50).all()
    if not items:
        items = []
    return items



from datetime import date
from typing import Any, List

from pydantic import conlist
from pydantic.utils import GetterDict

from mavedb.view_models import external_gene_identifier_offset
from mavedb.view_models.base.base import BaseModel, validator
from mavedb.view_models.reference_map import ReferenceMap, ReferenceMapCreate
from mavedb.view_models.wild_type_sequence import WildTypeSequence, WildTypeSequenceCreate
from mavedb.lib.validation import target


class ExternalIdentifiersGetter(GetterDict):
    """
    Custom class used in transforming TargetGene SQLAlchemy model objects into Pydantic view model objects, with special
    handling of external identifiers for target genes.

    Pydantic uses GetterDict objects to access source objects as dictionaries, which can then be turned into Pydantic
    view model objects. Here we need to remap the underlying SQLAlchemy model's separate properties for Ensembl, RefSeq,
    and UniProt identifiers into an array of external identifiers.
    """

    def get(self, key: str, default: Any) -> Any:
        if key == "external_identifiers":
            ensembl_offset = getattr(self._obj, "ensembl_offset")
            refseq_offset = getattr(self._obj, "refseq_offset")
            uniprot_offset = getattr(self._obj, "uniprot_offset")
            return list(filter(lambda x: x is not None, [ensembl_offset, refseq_offset, uniprot_offset]))
        else:
            return super().get(key, default)


class TargetGeneBase(BaseModel):
    """Base class for target gene view models."""

    name: str
    category: str
    external_identifiers: list[external_gene_identifier_offset.ExternalGeneIdentifierOffsetBase]

    class Config:
        getter_dict: ExternalIdentifiersGetter


class TargetGeneModify(TargetGeneBase):
    @validator("category")
    def validate_category(cls, v):
        target.validate_target_category(v)
        return v


class TargetGeneCreate(TargetGeneModify):
    """View model for creating a new target gene."""

    reference_maps: conlist(ReferenceMapCreate, min_items=1)
    wt_sequence: WildTypeSequenceCreate
    external_identifiers: list[external_gene_identifier_offset.ExternalGeneIdentifierOffsetCreate]


class TargetGeneUpdate(TargetGeneModify):
    """View model for updating a target gene."""

    external_identifiers: list[external_gene_identifier_offset.ExternalGeneIdentifierOffsetCreate]


class SavedTargetGene(TargetGeneBase):
    """Base class for target gene view models representing saved records."""

    external_identifiers: list[external_gene_identifier_offset.SavedExternalGeneIdentifierOffset]

    class Config:
        orm_mode = True
        arbitrary_types_allowed = True


class TargetGene(SavedTargetGene):
    """Target gene view model containing a complete set of properties visible to non-admin users."""

    reference_maps: List[ReferenceMap]
    wt_sequence: WildTypeSequence
    external_identifiers: list[external_gene_identifier_offset.ExternalGeneIdentifierOffset]

    class Config:
        getter_dict = ExternalIdentifiersGetter


class ShortTargetGene(SavedTargetGene):
    """Target gene view model containing a smaller set of properties to return in list contexts."""

    reference_maps: List[ReferenceMap]
    external_identifiers: list[external_gene_identifier_offset.ExternalGeneIdentifierOffset]

    class Config:
        getter_dict = ExternalIdentifiersGetter


class AdminTargetGene(SavedTargetGene):
    """Target gene view model containing properties to return to admin clients."""

    creation_date: date
    modification_date: date
    reference_maps: List[ReferenceMap]
    wt_sequence: WildTypeSequence
    external_identifiers: list[external_gene_identifier_offset.ExternalGeneIdentifierOffset]

    class Config:
        getter_dict = ExternalIdentifiersGetter



from typing import Optional

from mavedb.view_models.base.base import BaseModel, validator
from mavedb.lib.validation import identifier as identifier_validator


class ExternalGeneIdentifierBase(BaseModel):
    db_name: str
    identifier: str


class ExternalGeneIdentifierCreate(ExternalGeneIdentifierBase):
    @validator("db_name")
    def validate_db_name(cls, value):
        identifier_validator.validate_db_name(value)
        return value

    @validator("identifier")
    def validate_identifier(cls, field_value, values, field, config):
        # if db_name is none, values["db_name"] will raise an error.
        if "db_name" in values.keys():
            db_name = values["db_name"]
            # field_value is identifier
            identifier_validator.validate_identifier(db_name, field_value)
        return field_value


# Properties shared by models stored in DB
class SavedExternalGeneIdentifier(ExternalGeneIdentifierBase):
    db_version: Optional[str]
    url: Optional[str]
    reference_html: Optional[str]

    class Config:
        orm_mode = True


# Properties to return to non-admin clients
class ExternalGeneIdentifier(SavedExternalGeneIdentifier):
    pass



# See https://pydantic-docs.helpmanual.io/usage/postponed_annotations/#self-referencing-models
from __future__ import annotations
from datetime import date
from typing import Dict, Optional

from mavedb.view_models.base.base import BaseModel, validator
from mavedb.view_models.doi_identifier import DoiIdentifier, DoiIdentifierCreate, SavedDoiIdentifier
from mavedb.view_models.experiment import Experiment, SavedExperiment
from mavedb.view_models.license import License, SavedLicense, ShortLicense
from mavedb.view_models.pubmed_identifier import PubmedIdentifier, PubmedIdentifierCreate, SavedPubmedIdentifier
from mavedb.view_models.target_gene import SavedTargetGene, ShortTargetGene, TargetGene, TargetGeneCreate
from mavedb.view_models.user import SavedUser, User
from mavedb.view_models.variant import VariantInDbBase
from mavedb.lib.validation import keywords
from mavedb.lib.validation import urn


class ScoresetBase(BaseModel):
    """Base class for score set view models."""

    title: str
    method_text: Optional[str]
    abstract_text: Optional[str]
    short_description: str
    extra_metadata: Dict
    data_usage_policy: Optional[str]
    keywords: Optional[list[str]]


class ScoresetModify(ScoresetBase):
    @validator("keywords")
    def validate_keywords(cls, v):
        keywords.validate_keywords(v)
        return v


class ScoresetCreate(ScoresetModify):
    """View model for creating a new score set."""

    experiment_urn: str
    license_id: int
    superseded_scoreset_urn: Optional[str]
    meta_analysis_source_scoreset_urns: Optional[list[str]]
    target_gene: TargetGeneCreate
    doi_identifiers: Optional[list[DoiIdentifierCreate]]
    pubmed_identifiers: Optional[list[PubmedIdentifierCreate]]

    @validator("superseded_scoreset_urn", "meta_analysis_source_scoreset_urns")
    def validate_scoreset_urn(cls, v):
        if v is None:
            pass
        # For superseded_scoreset_urn
        elif type(v) == str:
            urn.validate_mavedb_urn_scoreset(v)
        # For meta_analysis_source_scoreset_urns
        else:
            [urn.validate_mavedb_urn_scoreset(s) for s in v]
        return v

    @validator("experiment_urn")
    def validate_experiment_urn(cls, v):
        urn.validate_mavedb_urn_experiment(v)
        return v


class ScoresetUpdate(ScoresetModify):
    """View model for updating a score set."""

    license_id: Optional[int]
    doi_identifiers: list[DoiIdentifierCreate]
    pubmed_identifiers: list[PubmedIdentifierCreate]
    target_gene: TargetGeneCreate


class ShortScoreset(BaseModel):
    """
    Score set view model containing a smaller set of properties to return in list contexts.

    Notice that this is not derived from ScoresetBase.
    """

    urn: str
    title: str
    short_description: str
    published_date: Optional[date]
    replaces_id: Optional[int]
    num_variants: int
    experiment: Experiment
    license: ShortLicense
    creation_date: date
    modification_date: date
    target_gene: ShortTargetGene
    private: bool

    class Config:
        orm_mode = True
        arbitrary_types_allowed = True


class SavedScoreset(ScoresetBase):
    """Base class for score set view models representing saved records."""

    urn: str
    num_variants: int
    experiment: SavedExperiment
    license: ShortLicense
    superseded_scoreset: Optional[ShortScoreset]
    superseding_scoreset: Optional[SavedScoreset]
    meta_analysis_source_scoresets: list[ShortScoreset]
    meta_analyses: list[ShortScoreset]
    doi_identifiers: list[SavedDoiIdentifier]
    pubmed_identifiers: list[SavedPubmedIdentifier]
    published_date: Optional[date]
    creation_date: date
    modification_date: date
    created_by: Optional[SavedUser]
    modified_by: Optional[SavedUser]
    target_gene: SavedTargetGene
    dataset_columns: Dict

    class Config:
        orm_mode = True
        arbitrary_types_allowed = True


class Scoreset(SavedScoreset):
    """Score set view model containing most properties visible to non-admin users, but no variant data."""

    experiment: Experiment
    license: ShortLicense
    superseded_scoreset: Optional[ShortScoreset]
    superseding_scoreset: Optional[Scoreset]
    meta_analysis_source_scoresets: list[ShortScoreset]
    meta_analyses: list[ShortScoreset]
    doi_identifiers: list[DoiIdentifier]
    pubmed_identifiers: list[PubmedIdentifier]
    created_by: Optional[User]
    modified_by: Optional[User]
    target_gene: TargetGene
    num_variants: int
    private: bool
    # processing_state: Optional[str]


class ScoresetWithVariants(Scoreset):
    """
    Score set view model containing a complete set of properties visible to non-admin users, for contexts where variants
    are requested.
    """

    variants: list[VariantInDbBase]


class AdminScoreset(Scoreset):
    """Score set view model containing properties to return to admin clients."""

    normalised: bool
    approved: bool



from datetime import date

from sqlalchemy import Column, Date, Integer, String

from mavedb.db.base import Base


class RawReadIdentifier(Base):
    __tablename__ = "sra_identifiers"

    id = Column(Integer, primary_key=True, index=True)
    identifier = Column(String, nullable=False)
    db_name = Column(String, nullable=False)
    db_version = Column(String, nullable=True)
    url = Column(String, nullable=True)
    creation_date = Column(Date, nullable=False, default=date.today)
    modification_date = Column(Date, nullable=False, default=date.today, onupdate=date.today)



from mavedb.view_models.base.base import BaseModel, validator
from mavedb.view_models import external_gene_identifier


class ExternalGeneIdentifierOffsetBase(BaseModel):
    identifier: external_gene_identifier.ExternalGeneIdentifierBase
    offset: int


class ExternalGeneIdentifierOffsetCreate(ExternalGeneIdentifierOffsetBase):
    identifier: external_gene_identifier.ExternalGeneIdentifierCreate

    @validator("offset")
    def validate_offset(cls, v):
        if v < 0:
            raise ValueError("Offset should not be a negative number")
        return v


# Properties shared by models stored in DB
class SavedExternalGeneIdentifierOffset(ExternalGeneIdentifierOffsetBase):
    identifier: external_gene_identifier.SavedExternalGeneIdentifier

    class Config:
        orm_mode = True


# Properties to return to non-admin clients
class ExternalGeneIdentifierOffset(SavedExternalGeneIdentifierOffset):
    identifier: external_gene_identifier.ExternalGeneIdentifier



from sqlalchemy import Column, Integer, String

from mavedb.db.base import Base


class DjangoReversionVersion(Base):
    __tablename__ = "reversion_version"

    id = Column(Integer, primary_key=True, index=True)
    object_id = Column(String(191), nullable=False)
    format = Column(String(255), nullable=False)
    serialized_data = Column(String, nullable=False)
    object_repr = Column(String, nullable=False)
    content_type_id = Column(Integer, nullable=False)
    revision_id = Column(Integer, nullable=False)
    db = Column(String(191), nullable=True)



from sqlalchemy import Column, DateTime, Integer, String

from mavedb.db.base import Base


class DjangoReversionRevision(Base):
    __tablename__ = "reversion_revision"

    id = Column(Integer, primary_key=True, index=True)
    date_created = Column(DateTime(timestamp=True), nullable=False)
    comment = Column(String, nullable=False)
    user_id = Column(Integer, nullable=True)



from typing import Optional, Union

from mavehgvs import Variant, MaveHgvsParseError
from mavedb.lib.validation.exceptions import ValidationError
from mavedb.lib.validation.constants.general import (
    hgvs_nt_column,
    hgvs_splice_column,
    hgvs_pro_column,
)

__all__ = ["validate_hgvs_string"]

from mavedb.lib.validation.utilities import is_null


def validate_hgvs_string(
    value: Union[str, bytes],
    column: Optional[str] = None,
    splice_present: bool = False,
    targetseq: Optional[str] = None,
    relaxed_ordering: bool = False,
) -> Optional[str]:
    """
    Validates hgvs string.

    Parameters
    __________
    value : Union[str, bytes]
    column : Optional[str] = None
    splice_present : bool = False
    targetseq : Optional[str] = None
    relaxed_ordering : bool = False

    Returns
    _______
    Optional[str]

    Raises
    ______
    ValidationError
        If variant HGVS input values are not strings.
    ValidationError
        If value is _sy or _wt, which are no longer supported.
    ValidationError
        If
    ValidationError
        If value is not a genomic variant (prefix 'g.'). Nucleotide variants must
        be genomic if transcript variants are also defined.
    ValidationError
        If value is not a transcript variant. The accepted transcript variant
        prefixes are 'c.', 'n.'.
    ValidationError
        If value is not a protein variant. The accepted protein variant prefix is 'p.'.
    ValueError
        If there exists an unknown column. Function expects nt, splice or p."
    """
    if is_null(value):
        return None

    if hasattr(value, "decode"):
        value = value.decode()
    if not isinstance(value, str):
        raise ValidationError(
            "Variant HGVS values input must be strings. " "'{}' has the type '{}'.".format(value, type(value).__name__)
        )

    if value.lower() == "_sy":
        raise ValidationError("_sy is no longer supported and should be replaced by p.(=)")
    elif value.lower() == "_wt":
        raise ValidationError("_wt is no longer supported and should be replaced by (cgnp).=")

    try:
        variant = Variant(s=value, targetseq=targetseq, relaxed_ordering=relaxed_ordering)
    except MaveHgvsParseError as error:
        raise ValidationError(f"{value}: {str(error)}")

    prefix = variant.prefix.lower()
    if column in ("nt", hgvs_nt_column):
        if splice_present:
            if prefix not in "g":
                raise ValidationError(
                    f"'{value}' is not a genomic variant (prefix 'g.'). "
                    f"Nucleotide variants must be genomic if transcript "
                    f"variants are also defined."
                )
        else:
            if prefix not in "cn":
                raise ValidationError(
                    f"'{value}' is not a transcript variant. The accepted "
                    f"transcript variant prefixes are 'c.', 'n.'."
                )
    elif column in ("splice", hgvs_splice_column):
        if prefix not in "cn":
            raise ValidationError(
                f"'{value}' is not a transcript variant. The accepted " f"transcript variant prefixes are 'c.', 'n.'."
            )
    elif column in ("p", hgvs_pro_column):
        if prefix not in "p":
            raise ValidationError(
                f"'{value}' is not a protein variant. The accepted " f"protein variant prefix is 'p.'."
            )
    else:
        raise ValueError("Unknown column '{}'. Expected nt, splice or p".format(column))

    return str(variant)


# validate_nt_variant = partial(validate_hgvs_string, **{"column": "nt"})
# validate_splice_variant = partial(validate_hgvs_string, **{"column": "splice"})
# validate_pro_variant = partial(validate_hgvs_string, **{"column": "p"})



from unittest import TestCase

from mavedb.lib.validation.constants.general import null_values_list
from mavedb.lib.validation.variant import validate_hgvs_string  # validate_pro_variant, validate_nt_variant

from mavedb.lib.validation.utilities import is_null, generate_hgvs, construct_hgvs_pro


class TestIsNull(TestCase):
    def test_valid_null_values(self):
        for value in null_values_list:
            self.assertTrue(is_null(value))

    def test_invalid_null_values(self):
        self.assertFalse(is_null(1))
        self.assertFalse(is_null("1"))


class TestGenerateHgvsPro(TestCase):
    def test_pro(self):
        pro = generate_hgvs("p")
        validate_hgvs_string(pro)

    def test_nt(self):
        nt = generate_hgvs()
        validate_hgvs_string(nt)


class TestConstructHgvsPro(TestCase):
    def test_valid_arguments(self):
        construct_hgvs_pro(wt="Ala", mutant="Gly", position=3)

    def test_invalid_wt_aa(self):
        with self.assertRaises(ValueError):
            construct_hgvs_pro(wt="Alr", mutant="Gly", position=3)

    def test_invalid_mut_aa(self):
        with self.assertRaises(ValueError):
            construct_hgvs_pro(wt="Ala", mutant="Gla", position=3)

    def test_invalid_position(self):
        # TODO what are the invalid positions we should consider?
        self.assertFalse(False)


class TestConvertHgvsNtToHgvsPro(TestCase):
    def setUp(self):
        self.target_seq = "ATGACA"
        self.hgvs_nt_values = ["g.4A>G", "g.5C>G", "g.6A>G"]
        self.hgvs_pro_values = ["p.Thr2Ala", "p.Thr2Arg", "p.Thr2="]

    def test_wt_hgvs_nt(self):
        # convert_hgvs_nt_to_hgvs_pro(hgvs_nt="g.4A>G", )
        pass

    def test_wt_hgvs_pro(self):
        pass

    def test_deletion_hgvs_nt(self):
        pass

    def test_one_base_change_codon_variant(self):
        pass

    def test_two_base_change_codon_variant(self):
        pass

    def test_three_base_change_codon_variant(self):
        pass


class TestVariantTypeHelperFunctions(TestCase):
    def test_test_is_wild_type(self):
        pass

    def test_is_deletion(self):
        pass

    def test_test_is_substitution_one_base(self):
        pass

    def test_test_is_substitution_two_bases_nonadjacent(self):
        pass



import idutils

from mavedb.view_models.base.base import BaseModel, validator
from mavedb.lib.validation.exceptions import ValidationError


class DoiIdentifierBase(BaseModel):
    identifier: str


class DoiIdentifierCreate(DoiIdentifierBase):
    @validator("identifier")
    def must_be_valid_doi(cls, v):
        if not idutils.is_doi(v):
            raise ValidationError("{} is not a valid DOI identifier.".format(v))
        return v


# Properties shared by models stored in DB
class SavedDoiIdentifier(DoiIdentifierBase):
    id: int
    url: str

    class Config:
        orm_mode = True


# Properties to return to non-admin clients
class DoiIdentifier(SavedDoiIdentifier):
    pass



from datetime import date

from sqlalchemy import Column, Date, Integer, String

from mavedb.db.base import Base


class Keyword(Base):
    __tablename__ = "keywords"

    id = Column(Integer, primary_key=True, index=True)
    text = Column(String, nullable=False, unique=True)
    creation_date = Column(Date, nullable=False, default=date.today)
    modification_date = Column(Date, nullable=False, default=date.today, onupdate=date.today)



from datetime import date

from sqlalchemy import Column, Date, ForeignKey, Integer, String
from sqlalchemy.orm import relationship

from mavedb.db.base import Base
from .genome_identifier import GenomeIdentifier


class ReferenceGenome(Base):
    __tablename__ = "reference_genomes"

    id = Column(Integer, primary_key=True, index=True)
    short_name = Column(String, nullable=False)
    organism_name = Column(String, nullable=False)
    genome_identifier_id = Column(Integer, ForeignKey("genome_identifiers.id"), nullable=True)
    genome_identifier = relationship("GenomeIdentifier", backref="reference_genomes")
    creation_date = Column(Date, nullable=False, default=date.today)
    modification_date = Column(Date, nullable=False, default=date.today, onupdate=date.today)



import csv
import io
import logging
import re
from datetime import date
from typing import Any, List, Optional

import numpy as np
import pandas as pd
from fastapi import APIRouter, Depends, File, status, UploadFile
from fastapi.encoders import jsonable_encoder
from fastapi.exceptions import HTTPException
from fastapi.responses import StreamingResponse
from sqlalchemy.orm import Session
from sqlalchemy.exc import MultipleResultsFound

from mavedb import deps
from mavedb.lib.authorization import require_current_user
from mavedb.lib.identifiers import (
    create_external_gene_identifier_offset,
    find_or_create_doi_identifier,
    find_or_create_pubmed_identifier,
)
from mavedb.lib.scoresets import create_variants_data, search_scoresets as _search_scoresets, VariantData
from mavedb.lib.urns import generate_experiment_set_urn, generate_experiment_urn, generate_scoreset_urn
from mavedb.lib.validation import exceptions
from mavedb.lib.validation.constants.general import null_values_list
from mavedb.lib.validation.dataframe import validate_and_standardize_dataframe_pair
from mavedb.models.enums.processing_state import ProcessingState
from mavedb.models.experiment import Experiment
from mavedb.models.license import License
from mavedb.models.reference_map import ReferenceMap
from mavedb.models.scoreset import Scoreset
from mavedb.models.target_gene import TargetGene
from mavedb.models.user import User
from mavedb.models.variant import Variant
from mavedb.models.wild_type_sequence import WildTypeSequence
from mavedb.view_models import scoreset
from mavedb.view_models.search import ScoresetsSearch

logger = logging.getLogger(__name__)

null_values_re = re.compile(r"\s+|none|nan|na|undefined|n/a|null|nil", flags=re.IGNORECASE)


async def fetch_scoreset_by_urn(db, urn):
    try:
        # item = db.query(Scoreset).filter(Scoreset.urn == urn).filter(Scoreset.private.is_(False)).one_or_none()
        item = db.query(Scoreset).filter(Scoreset.urn == urn).one_or_none()
    except MultipleResultsFound:
        raise HTTPException(status_code=500, detail=f"Multiple scoresets with URN {urn} were found.")
    if not item:
        raise HTTPException(status_code=404, detail=f"Scoreset with URN {urn} not found")
    return item


def is_null(value):
    """Return True if a string represents a null value."""
    value = str(value).strip().lower()
    return null_values_re.fullmatch(value) or not value


router = APIRouter(prefix="/api/v1", tags=["scoresets"], responses={404: {"description": "Not found"}})


@router.post("/scoresets/search", status_code=200, response_model=list[scoreset.ShortScoreset])
def search_scoresets(search: ScoresetsSearch, db: Session = Depends(deps.get_db)) -> Any:  # = Body(..., embed=True),
    """
    Search scoresets.
    """
    return _search_scoresets(db, None, search)


@router.post("/me/scoresets/search", status_code=200, response_model=list[scoreset.ShortScoreset])
def search_my_scoresets(
    search: ScoresetsSearch,  # = Body(..., embed=True),
    db: Session = Depends(deps.get_db),
    user: User = Depends(require_current_user),
) -> Any:
    """
    Search scoresets created by the current user..
    """
    return _search_scoresets(db, user, search)


@router.get("/scoresets/{urn}", status_code=200, response_model=scoreset.Scoreset, responses={404: {}, 500: {}})
async def show_scoreset(*, urn: str, db: Session = Depends(deps.get_db)) -> Any:
    """
    Fetch a single scoreset by URN.
    """

    return await fetch_scoreset_by_urn(db, urn)


@router.get(
    "/scoresets/{urn}/scores",
    status_code=200,
    responses={
        200: {
            "content": {"text/csv": {}},
            "description": """Variant scores in CSV format, with four fixed columns (accession, hgvs_nt, hgvs_pro,"""
            """ and hgvs_splice), plus score columns defined by the scoreset.""",
        }
    },
)
def get_scoreset_scores_csv(
    *,
    urn: str,
    db: Session = Depends(deps.get_db),
) -> Any:
    """
    Return scores from a scoreset, identified by URN, in CSV format.
    """
    scoreset = db.query(Scoreset).filter(Scoreset.urn == urn).first()
    if not scoreset:
        raise HTTPException(status_code=404, detail=f"Scoreset with URN {urn} not found")
    columns = ["accession", "hgvs_nt", "hgvs_splice", "hgvs_pro"] + scoreset.dataset_columns["score_columns"]
    type_column = "score_data"
    rows_data = get_csv_rows_data(scoreset.variants, columns=columns, dtype=type_column)
    stream = io.StringIO()
    writer = csv.DictWriter(stream, fieldnames=columns, quoting=csv.QUOTE_MINIMAL)
    writer.writeheader()
    writer.writerows(rows_data)
    return StreamingResponse(iter([stream.getvalue()]), media_type="text/csv")


@router.get(
    "/scoresets/{urn}/counts",
    status_code=200,
    responses={
        200: {
            "content": {"text/csv": {}},
            "description": """Variant counts in CSV format, with four fixed columns (accession, hgvs_nt, hgvs_pro,"""
            """ and hgvs_splice), plus score columns defined by the scoreset.""",
        }
    },
)
async def get_scoreset_counts_csv(
    *,
    urn: str,
    db: Session = Depends(deps.get_db),
) -> Any:
    """
    Return counts from a scoreset, identified by URN, in CSV format.
    """
    scoreset = db.query(Scoreset).filter(Scoreset.urn == urn).first()
    if not scoreset:
        raise HTTPException(status_code=404, detail=f"Recipe with ID {id} not found")
    columns = ["accession", "hgvs_nt", "hgvs_splice", "hgvs_pro"] + scoreset.dataset_columns["count_columns"]
    type_column = "count_data"
    rows_data = get_csv_rows_data(scoreset.variants, columns=columns, dtype=type_column)
    stream = io.StringIO()
    writer = csv.DictWriter(stream, fieldnames=columns, quoting=csv.QUOTE_MINIMAL)
    writer.writeheader()
    writer.writerows(rows_data)
    return StreamingResponse(iter([stream.getvalue()]), media_type="text/csv")


class HGVSColumns:
    NUCLEOTIDE: str = "hgvs_nt"  # dataset.constants.hgvs_nt_column
    TRANSCRIPT: str = "hgvs_splice"  # dataset.constants.hgvs_splice_column
    PROTEIN: str = "hgvs_pro"  # dataset.constants.hgvs_pro_column

    @classmethod
    def options(cls) -> List[str]:
        return [cls.NUCLEOTIDE, cls.TRANSCRIPT, cls.PROTEIN]


@router.post("/scoresets/", response_model=scoreset.Scoreset, responses={422: {}})
async def create_scoreset(
    *,
    item_create: scoreset.ScoresetCreate,
    db: Session = Depends(deps.get_db),
    user: User = Depends(require_current_user),
) -> Any:
    """
    Create a scoreset.
    """

    # TODO Confirm that the experiment is editable by this user.

    if item_create is None:
        return None

    experiment = db.query(Experiment).filter(Experiment.urn == item_create.experiment_urn).one_or_none()
    if not experiment:
        raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail="Unknown experiment")

    license_ = db.query(License).filter(License.id == item_create.license_id).one_or_none()
    if not license_:
        raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail="Unknown license")

    if item_create.superseded_scoreset_urn is not None:
        superseded_scoreset = await fetch_scoreset_by_urn(db, item_create.superseded_scoreset_urn)
    else:
        superseded_scoreset = None
    meta_analysis_source_scoresets = [
        await fetch_scoreset_by_urn(db, urn) for urn in item_create.meta_analysis_source_scoreset_urns or []
    ]

    doi_identifiers = [
        await find_or_create_doi_identifier(db, identifier.identifier)
        for identifier in item_create.doi_identifiers or []
    ]
    pubmed_identifiers = [
        await find_or_create_pubmed_identifier(db, identifier.identifier)
        for identifier in item_create.pubmed_identifiers or []
    ]
    wt_sequence = WildTypeSequence(**jsonable_encoder(item_create.target_gene.wt_sequence, by_alias=False))
    target_gene = TargetGene(
        **jsonable_encoder(
            item_create.target_gene,
            by_alias=False,
            exclude=["external_identifiers", "reference_maps", "wt_sequence"],
        ),
        wt_sequence=wt_sequence,
    )
    for external_gene_identifier_offset_create in item_create.target_gene.external_identifiers:
        offset = external_gene_identifier_offset_create.offset
        identifier_create = external_gene_identifier_offset_create.identifier
        await create_external_gene_identifier_offset(
            db, target_gene, identifier_create.db_name, identifier_create.identifier, offset
        )
    reference_map = ReferenceMap(genome_id=item_create.target_gene.reference_maps[0].genome_id, target=target_gene)
    item = Scoreset(
        **jsonable_encoder(
            item_create,
            by_alias=False,
            exclude=[
                "doi_identifiers",
                "experiment_urn",
                "keywords",
                "license_id",
                "meta_analysis_source_scoreset_urns",
                "pubmed_identifiers",
                "superseded_scoreset_urn",
                "target_gene",
            ],
        ),
        experiment=experiment,
        license=license_,
        superseded_scoreset=superseded_scoreset,
        meta_analysis_source_scoresets=meta_analysis_source_scoresets,
        target_gene=target_gene,
        doi_identifiers=doi_identifiers,
        pubmed_identifiers=pubmed_identifiers,
        processing_state=ProcessingState.incomplete,
        created_by=user,
        modified_by=user,
    )
    await item.set_keywords(db, item_create.keywords)
    db.add(item)
    db.commit()
    db.refresh(item)
    return item


@router.post("/scoresets/{urn}/variants/data", response_model=scoreset.Scoreset, responses={422: {}})
async def upload_scoreset_variant_data(
    *,
    urn: str,
    counts_file: Optional[UploadFile] = File(None),
    scores_file: UploadFile = File(...),
    db: Session = Depends(deps.get_db),
    user: User = Depends(require_current_user),
) -> Any:
    """
    Upload scores and variant count files for a scoreset, and initiate processing these files to
    create variants.
    """

    # TODO Confirm access.

    # item = db.query(Scoreset).filter(Scoreset.urn == urn).filter(Scoreset.private.is_(False)).one_or_none()
    item = db.query(Scoreset).filter(Scoreset.urn == urn).one_or_none()
    if not item.urn or not scores_file:
        return None

    # Delete the old variants so that uploading new scores and counts won't accumulate the old ones.
    db.query(Variant).filter(Variant.scoreset_id == item.id).delete()

    extra_na_values = set(
        list(null_values_list)
        + [str(x).lower() for x in null_values_list]
        + [str(x).upper() for x in null_values_list]
        + [str(x).capitalize() for x in null_values_list]
    )
    scores_df = pd.read_csv(
        filepath_or_buffer=scores_file.file,
        sep=",",
        encoding="utf-8",
        quotechar="'",
        comment="#",
        na_values=extra_na_values,
        keep_default_na=True,
        dtype={**{col: str for col in HGVSColumns.options()}, "scores": float},
    )  # .replace(null_values_re, np.NaN) String will be replaced to NaN value
    for c in HGVSColumns.options():
        if c not in scores_df.columns:
            scores_df[c] = np.NaN
    score_columns = [col for col in scores_df.columns if col not in HGVSColumns.options()]
    counts_df = None
    count_columns = []
    if counts_file and counts_file.filename:
        counts_df = pd.read_csv(
            filepath_or_buffer=counts_file.file,
            sep=",",
            encoding="utf-8",
            quotechar="'",
            comment="#",
            na_values=extra_na_values,
            keep_default_na=True,
            dtype={**{col: str for col in HGVSColumns.options()}, "scores": float},
        ).replace(null_values_re, np.NaN)
        for c in HGVSColumns.options():
            if c not in counts_df.columns:
                counts_df[c] = np.NaN

        count_columns = [col for col in counts_df.columns if col not in HGVSColumns.options()]
    """
    if counts_file:
        try:
            validate_column_names(counts_df)
        except exceptions.ValidationError as e:
            raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail=str(e))
    """
    if scores_file:
        try:
            validate_and_standardize_dataframe_pair(
                scores_df, counts_df, item.target_gene.wt_sequence.sequence, item.target_gene.wt_sequence.sequence_type
            )
        except exceptions.ValidationError as e:
            raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail=str(e))

    variants_data = create_variants_data(scores_df, counts_df, None)  # , index_col)
    if variants_data:
        logger.error(f"{item.urn}:{variants_data[-1]}")

    item.num_variants = create_variants(db, item, variants_data)
    logger.error(f"Saving {item.urn}")

    item.dataset_columns = {"count_columns": list(count_columns), "score_columns": list(score_columns)}
    item.modified_by = user
    db.add(item)

    # create_variants_task.submit_task(kwargs={
    #    'user_id': None,
    #    'scoreset_urn': item.urn,
    #    'scores': None
    #    # 'counts',
    #    # 'index_col',
    #    # 'dataset_columns'
    # })

    db.commit()
    db.refresh(item)
    return item


# @classmethod
# @transaction.atomic
def create_variants(db, scoreset: Scoreset, variants_data: list[VariantData], batch_size=None) -> int:
    num_variants = len(variants_data)
    variant_urns = bulk_create_urns(num_variants, scoreset, True)
    variants = (Variant(urn=urn, scoreset_id=scoreset.id, **kwargs) for urn, kwargs in zip(variant_urns, variants_data))
    db.bulk_save_objects(variants)
    db.add(scoreset)
    return len(scoreset.variants)


# @staticmethod
def bulk_create_urns(n, scoreset, reset_counter=False) -> List[str]:
    start_value = 0 if reset_counter else scoreset.num_variants
    parent_urn = scoreset.urn
    child_urns = ["{}#{}".format(parent_urn, start_value + (i + 1)) for i in range(n)]
    current_value = start_value + n
    scoreset.num_variants = current_value
    return child_urns


@router.put("/scoresets/{urn}", response_model=scoreset.Scoreset, responses={422: {}})
async def update_scoreset(
    *,
    urn: str,
    item_update: scoreset.ScoresetUpdate,
    db: Session = Depends(deps.get_db),
    user: User = Depends(require_current_user),
) -> Any:
    """
    Update a scoreset .
    """

    if not item_update:
        raise HTTPException(status_code=400, detail="The request contained no updated item.")

    # item = db.query(Scoreset).filter(Scoreset.urn == urn).filter(Scoreset.private.is_(False)).one_or_none()
    item = db.query(Scoreset).filter(Scoreset.urn == urn).one_or_none()
    if not item:
        raise HTTPException(status_code=404, detail=f"Scoreset with URN {urn} not found.")
    # TODO Ensure that the current user has edit rights for this scoreset.

    # Editing unpublished scoreset
    if item.private is True:

        license_ = None
        if item_update.license_id is not None:
            license_ = db.query(License).filter(License.id == item_update.license_id).one_or_none()
            if not license_:
                raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail="Unknown license")
            item.license = license_

        for var, value in vars(item_update).items():
            if var not in ["keywords", "doi_identifiers", "experiment_urn", "license_id", "pubmed_identifiers", "target_gene"]:
                setattr(item, var, value) if value else None
        item.doi_identifiers = [
            await find_or_create_doi_identifier(db, identifier.identifier)
            for identifier in item_update.doi_identifiers or []
        ]

        item.pubmed_identifiers = [
            await find_or_create_pubmed_identifier(db, identifier.identifier)
            for identifier in item_update.pubmed_identifiers or []
        ]

        await item.set_keywords(db, item_update.keywords)

        # Delete the old target gene, WT sequence, and reference map. These will be deleted when we set the scoreset's
        # target_gene to None, because we have set cascade='all,delete-orphan' on Scoreset.target_gene. (Since the
        # relationship is defined with the target gene as owner, this is actually set up in the backref attribute of
        # TargetGene.scoreset.)
        #
        # We must flush our database queries now so that the old target gene will be deleted before inserting a new one
        # with the same scoreset_id.
        item.target_gene = None
        db.flush()

        wt_sequence = WildTypeSequence(**jsonable_encoder(item_update.target_gene.wt_sequence, by_alias=False))

        target_gene = TargetGene(
            **jsonable_encoder(
                item_update.target_gene,
                by_alias=False,
                exclude=["external_identifiers", "reference_maps", "wt_sequence"],
            ),
            wt_sequence=wt_sequence,
        )
        for external_gene_identifier_offset_create in item_update.target_gene.external_identifiers:
            offset = external_gene_identifier_offset_create.offset
            identifier_create = external_gene_identifier_offset_create.identifier
            external_gene_identifier = await create_external_gene_identifier_offset(
                db, target_gene, identifier_create.db_name, identifier_create.identifier, offset
            )
        item.target_gene = target_gene

        reference_map = ReferenceMap(genome_id=item_update.target_gene.reference_maps[0].genome_id, target=target_gene)
        for var, value in vars(item_update).items():
            if var not in ["keywords", "doi_identifiers", "experiment_urn", "pubmed_identifiers", "target_gene"]:
                setattr(item, var, value) if value else None
    # Editing published scoreset
    else:
        for var, value in vars(item_update).items():
            if var in ["title", "method_text", "abstract_text", "short_description"]:
                setattr(item, var, value) if value else None
    db.add(item)

    db.commit()
    db.refresh(item)
    return item


@router.delete("/scoresets/{urn}", responses={422: {}})
async def delete_scoreset(
    *, urn: str, db: Session = Depends(deps.get_db), user: User = Depends(require_current_user)
) -> Any:
    """
    Delete a scoreset .

    Raises

    Returns
    _______
    Does not return anything
    string : HTTP code 200 successful but returning content
    or
    communitcate to client whether the operation succeeded
    204 if successful but not returning content - likely going with this
    """
    item = db.query(Scoreset).filter(Scoreset.urn == urn).one_or_none()
    if not item:
        raise HTTPException(status_code=404, detail=f"Scoreset with URN {urn} not found.")
    # TODO Ensure that the current user has edit rights for this scoreset.
    db.delete(item)
    db.commit()


@router.post("/scoresets/{urn}/publish", status_code=200, response_model=scoreset.Scoreset)
def publish_scoreset(
    *, urn: str, db: Session = Depends(deps.get_db), user: User = Depends(require_current_user)
) -> Any:
    """
    Publish a scoreset.
    """
    item: Scoreset = db.query(Scoreset).filter(Scoreset.urn == urn).one_or_none()
    if not item:
        raise HTTPException(status_code=404, detail=f"Scoreset with URN {urn} not found")
    # TODO Ensure that the current user has edit rights for this scoreset.
    if not item.experiment:
        raise HTTPException(
            status_code=500, detail="This scoreset does not belong to an experiment and cannot be published."
        )
    if not item.experiment.experiment_set:
        raise HTTPException(
            status_code=500,
            detail="This scoreset's experiment does not belong to an experiment set and cannot be published.",
        )

    published_date = date.today()

    if item.experiment.experiment_set.private or not item.experiment.experiment_set.published_date:
        item.experiment.experiment_set.urn = generate_experiment_set_urn(db)
        item.experiment.experiment_set.private = False
        item.experiment.experiment_set.published_date = published_date
        db.add(item.experiment.experiment_set)

    if item.experiment.private or not item.experiment.published_date:
        # TODO Pass experiment.is_meta_analysis instead of False
        item.experiment.urn = generate_experiment_urn(db, item.experiment.experiment_set, False)
        item.experiment.private = False
        item.experiment.published_date = published_date
        db.add(item.experiment)

    item.urn = generate_scoreset_urn(db, item.experiment)
    item.private = False
    item.published_date = published_date
    db.add(item)

    db.commit()
    db.refresh(item)
    return item


def get_csv_rows_data(variants, columns, dtype, na_rep="NA"):
    """
    Format each variant into a dictionary row containing the keys specified
    in `columns`.

    Parameters
    ----------
    variants : list[variant.models.Variant`]
        List of variants.
    columns : list[str]
        Columns to serialize.
    dtype : str, {'scores', 'counts'}
        The type of data requested. Either the 'score_data' or 'count_data'.
    na_rep : str
        String to represent null values.

    Returns
    -------
    list[dict]
    """
    row_dicts = []
    for variant in variants:
        data = {}
        for column_key in columns:
            if column_key == "hgvs_nt":
                value = str(variant.hgvs_nt)
            elif column_key == "hgvs_pro":
                value = str(variant.hgvs_pro)
            elif column_key == "hgvs_splice":
                value = str(variant.hgvs_splice)
            elif column_key == "accession":
                value = str(variant.urn)
            else:
                value = str(variant.data[dtype][column_key])
            if is_null(value):
                value = na_rep
            data[column_key] = value
        row_dicts.append(data)
    return row_dicts



from mavedb.lib.validation.exceptions import ValidationError
from mavedb.lib.validation.utilities import is_null


def validate_keywords(keywords: list[str]):
    """
    Validates a list of keywords.

    Parameters
    __________
    keywords: list[str]
        A list of keywords.

    Raises
    ______
    ValidationError
        If the list is invalid or null or if any individual keyword is invalid or null.
    """
    if is_null(keywords):
        raise ValidationError(
            "{} are not valid keywords. Keywords must be a non null list of strings.".format(keywords)
        )
    else:
        for keyword in keywords:
            validate_keyword(keyword)


def validate_keyword(keyword: str):
    """
    This function validates whether or not the kw parameter is valid by
    checking that it is a string that is not null. If kw is null
    or is not a string, an error is raised.

    Parameters
    __________
    kw : str
        The keyword to be validated.

    Raises
    ______
    ValidationError
        If the kw argument is not a valid string.
    """
    if is_null(keyword) or not isinstance(keyword, str):
        raise ValidationError("{} not a valid keyword. Keywords must be non null strings.".format(keyword))



from random import choice
from typing import Optional

from mavedb.lib.validation.constants.conversion import codon_dict_DNA
from mavedb.lib.validation.constants.conversion import aa_dict_key_1
from mavedb.lib.validation.constants.general import null_values_re
from mavehgvs.variant import Variant


def is_null(value):
    """
    Returns True if a stripped/lowercase value in in `nan_col_values`.

    Parameters
    __________
    value : str
        The value to be checked as null or not.

    Returns
    _______
    bool
        True value is NoneType or if value matches the stated regex patterns in constants.null_values_re.
    """
    value = str(value).strip().lower()
    if not value:
        return True
    match = null_values_re.fullmatch(value)
    if match:
        return True
    else:
        return False
    # return null_values_re.fullmatch(value) or not value


def generate_hgvs(prefix: str = "c") -> str:
    """
    Generates a random hgvs string from a small sample.
    """
    if prefix == "p":
        # Subset of 3-letter codes, chosen at random.
        amino_acids = [
            "Ala",
            "Leu",
            "Gly",
            "Val",
            "Tyr",
            "Met",
            "Cys",
            "His",
            "Glu",
            "Phe",
        ]
        ref = choice(amino_acids)
        alt = choice(amino_acids)
        return f"{prefix}.{ref}{choice(range(1, 100))}{alt}"
    else:
        alt = choice("ATCG")
        ref = choice("ATCG")
        return f"{prefix}.{choice(range(1, 100))}{ref}>{alt}"


def construct_hgvs_pro(wt: str, mutant: str, position: int, target_seq: Optional[str] = None):
    # TODO: the testing on this function needs to be improved
    """
    Given the wt and mutant 3 lette amino acid codes as well as the position, this function generates a validated
    hgvs_pro string.

    Parameters
    __________
    wt: str
        The wt 3 letter amino acid code.
    mutant: str
        The mutant 3 letter amino acid code.
    position: int
        The position of the change.

    Returns
    _______
    hgvs
        The constructed hgvs_pro string.

    Raises
    ______
    ValueError
        If the wt or mutant 3 letter amino acid codes are invalid.
    """
    # TODO account for when variant codon is None, a deletion event
    # check that the provided 3 letter amino acid codes are valid
    if wt not in aa_dict_key_1.values():
        raise ValueError(
            "wt 3 letter amino acid code {} is invalid, "
            "must be one of the following: {}".format(wt, list(aa_dict_key_1.values()))
        )
    if mutant not in aa_dict_key_1.values():
        raise ValueError(
            "wt 3 letter amino acid code {} is invalid, "
            "must be one of the following: {}".format(mutant, list(aa_dict_key_1.values()))
        )

    if wt == mutant:
        hgvs = "p." + wt + str(position) + "="
    else:
        hgvs = "p." + wt + str(position) + mutant
    # validate variant
    Variant(hgvs)
    # var.validate_hgvs_string(value=hgvs, column="p", targetseq=target_seq)
    return hgvs


def convert_hgvs_nt_to_hgvs_pro(hgvs_nt: str, target_seq: str):
    # TODO note that this only works for codon changes and single mutants
    """
    This function takes a hgvs_nt variant string and its associated target sequence and returns
    a validated hgvs_pro equivalent.

    Parameters
    __________
    hgvs_nt: string
        The hgvs_nt string that will be converted.
    target_seq:
        The target sequence associated with the hgvs_nt variant.

    Raises
    ______
    TypeError
        If target_seq is not string.
    ValueError
        If target_seq is not made solely of characters ACTG.
    """
    # check that the hgvs_nt variant is valid with regards to the target sequence
    # validate_hgvs_string(value=hgvs_nt,
    #                     column="nt",
    #                     targetseq=target_seq)

    # check for TypeError
    # if target_seq is not string
    if not isinstance(target_seq, str):
        raise TypeError("target_seq must be string.")

    # check for ValueError
    # if target_seq is not made solely of characters ACTG
    check_chars = [letter in "ACTG" for letter in target_seq]
    if False in check_chars:
        raise ValueError("target_seq is invalid, must be composed only of bases ACTG.")

    # identify variant_position and get codon_number associated with it

    if _is_wild_type(hgvs_nt):  # variant_codon is wild-type
        codon_number = None
        target_codon = None
    else:  # any other variant change
        # instantiate Variant object
        variant = Variant(hgvs_nt)
        # get variant position and convert to int
        if type(variant.positions) == list:  # multiple positions values exist
            variant_position = int(str(variant.positions[0]))
        elif type(variant.positions) == tuple:
            variant_position = int(str(variant.positions[0]))
        else:  # only one value for positions
            variant_position = int(str(variant.positions))
        # now that we have the variant_position, get codon_number
        codon_number = round((variant_position / 3) + 0.5)
        # use codon_number to get target_codon from target_seq
        target_codon = target_seq[(codon_number - 1) * 3 : codon_number * 3]

    # declare variables for codon data
    # keep track of the number and location of the changes within the codon
    sub_one = None
    sub_two = None
    sub_three = None
    # keep tack of the number and value of the changes within the codon
    sub_one_nuc = None
    sub_two_nuc = None
    sub_three_nuc = None
    # keep track of the full codon changes
    variant_codon = None

    # determine sequence of variant_codon

    if _is_wild_type(hgvs_nt):  # variant_codon is wild-type
        variant_codon = target_codon
        sub_one = None  # no nucleotide substitutions
    elif _is_deletion(hgvs_nt):  # target_codon was deleted
        variant_codon = None
        sub_one = None  # no nucleotide substitutions
    elif _is_substitution_one_base(hgvs_nt):  # variant_codon has one nucleotide substitution
        # instantiate Variant object
        variant = Variant(hgvs_nt)
        # get index of nucleotide substitution
        sub_one = int(str(variant.positions)) % 3 - 1
        # get nucleotide of substitution
        sub_one_nuc = variant.sequence[1]
        # set other possible indices for codon substitution to None
        sub_two = None
        sub_three = None
    elif _is_substitution_two_bases_nonadjacent(hgvs_nt):  # variant has two nucleotide substitutions, non-adjacent
        # instantiate Variant object
        variant = Variant(hgvs_nt)
        # get indices of nucleotide substitutions
        sub_one = int(str(variant.positions[0])) % 3 - 1
        sub_two = int(str(variant.positions[1])) % 3 - 1
        # get nucleotides of substitutions
        sub_one_nuc = variant.sequence[0][1]
        sub_two_nuc = variant.sequence[1][1]
        # set other possible indices for codon substitution to None
        sub_three = None
    else:  # variant_codon has two or three adjacent nucleotide substitutions
        # instantiate Variant object
        variant = Variant(hgvs_nt)
        variant_codon = variant.sequence
        # get index of first codon substitution
        sub_one = int(str(variant.positions[0])) % 3 - 1
        # get string of substituted nucleotides
        sub_nucs = variant.sequence
        if len(sub_nucs) == 2:  # variant codon has two adjacent nucleotide substitutions
            # assign additional nucleotide substitution indices
            sub_two = sub_one + 1
            # get nucleotides of substitutions
            sub_one_nuc = sub_nucs[0]
            sub_two_nuc = sub_nucs[1]
            # set other possible indices for codon substitution to None
            sub_three = None
        else:  # variant has three adjacent nucleotide substitutions
            # assign additional nucleotide substitution indices
            sub_two = sub_one + 1
            sub_three = sub_two + 1
            # get nucleotides of substitutions
            sub_one_nuc = sub_nucs[0]
            sub_two_nuc = sub_nucs[1]
            sub_three_nuc = sub_nucs[2]

    # using data generated above (substituted nucleotides and indices in codon), construct variant_codon

    # only assign variant_codon if nucleotide substitution occurred
    if sub_one is not None:
        # declare and initialize variant_codon
        variant_codon = ""
        # set first nucleotide of variant_codon
        if sub_one == 0:
            variant_codon = variant_codon + sub_one_nuc
        else:
            variant_codon = variant_codon + target_codon[0]
        # set second nucleotide of variant_codon
        if sub_one == 1:
            variant_codon = variant_codon + sub_one_nuc
        elif sub_two == 1:
            variant_codon = variant_codon + sub_two_nuc
        else:
            variant_codon = variant_codon + target_codon[1]
        # set third nucleotide of variant_codon
        if sub_one == -1 or sub_one == 2:
            variant_codon = variant_codon + sub_one_nuc
        elif sub_two == -1 or sub_two == 2:
            variant_codon = variant_codon + sub_two_nuc
        elif sub_three == -1 or sub_three == 2:
            variant_codon = variant_codon + sub_three_nuc
        else:
            variant_codon = variant_codon + target_codon[2]

    # convert to 3 letter amino acid code
    target_aa = codon_dict_DNA[target_codon]
    if variant_codon:
        variant_aa = codon_dict_DNA[variant_codon]
    else:
        variant_aa = None

    return construct_hgvs_pro(wt=target_aa, mutant=variant_aa, position=codon_number, target_seq=target_seq)


def _is_wild_type(hgvs: str):
    # TODO this is no longer valid
    """
    This function takes an hgvs formatted string and returns True if the hgvs string indicates
    there was no change from the target sequence.

    Parameters
    ----------
    hgvs : string
        hgvs formatted string

    Returns
    -------
    wt : bool
        True if hgvs string indicates wild type
    """
    wt = False
    if hgvs.startswith("_wt"):
        wt = True
    return wt


def _is_deletion(hgvs: str):
    """
    This function takes an hgvs formatted string and returns True if the hgvs string indicates
    there was a deletion.

    Parameters
    ----------
    hgvs : string
        hgvs formatted string

    Returns
    -------
    deletion : bool
        True if hgvs string is indicates a deletion
    """
    deletion = False
    if hgvs.endswith("del"):
        deletion = True
    return deletion


def _is_substitution_one_base(hgvs: str):
    """
    This function takes an hgvs formatted string and returns True if the hgvs string indicates
    there was a substitution at one base of the codon.

    Parameters
    ----------
    hgvs : string
        hgvs formatted string

    Returns
    -------
    sub_one : bool
        True if hgvs string is indicates a substitution at one base of codon
    """
    sub_one = False
    if hgvs[-2] == ">":
        sub_one = True
    return sub_one


def _is_substitution_two_bases_nonadjacent(hgvs: str):
    """
    This function takes an hgvs formatted string and returns True if the hgvs string indicates
    there were substitutions (non-adjacent) in the codon.

    Parameters
    ----------
    hgvs : string
        hgvs formatted string

    Returns
    -------
    sub_two : bool
        True if hgvs string is indicates a substitution at one base of codon
    """
    sub_two = False
    if hgvs[-1] == "]":
        sub_two = True
    return sub_two
